# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
IsovarResult is a collection of all information gathered about a variant
and any protein sequences which were successfully translated for it.
"""

from __future__ import print_function, division, absolute_import

from collections import OrderedDict

from cached_property import cached_property

from .common import safediv
from .alignment_score import alignment_score

class IsovarResult(object):
    """
    This object represents all information gathered about a variant,
    which includes the AlleleReads supporting any allele at this variant's
    locus and any protein sequences generated from an alt-allele cDNA
    assembly.
    """

    def __init__(
            self,
            variant,
            read_evidence,
            predicted_effect,
            sorted_protein_sequences=None,
            filter_values=None,
            phased_variants_in_supporting_reads=None,
            phased_variants_in_protein_sequence=None):
        """
        Parameters
        ----------
        variant : varcode.Variant
            Somatic mutation.

        read_evidence : ReadEvidence
            Reads grouped by allele at the locus of this variant.

        predicted_effect : varcode effect
            Top predicted impact on protein sequence from DNA.

        sorted_protein_sequences : list of ProteinSequence
            Protein sequences sorted by degree of support.

        filter_values : OrderedDict
            Dictionary from names of filters to whether this IsovarResult
            passed that filter. 

        phased_variants_in_supporting_reads : set of varcode.Variant
            Other somatic variants which occur in the alt reads supporting the 
            variant associated with this IsovarResult.

        phased_variants_in_protein_sequence : set of varcode.Variant
            Other somatic variants which occur in the reads used to construct
            the top protein sequence associated with this IsovarResult.
        """
        self.variant = variant
        self.read_evidence = read_evidence
        self.predicted_effect = predicted_effect

        if sorted_protein_sequences is None:
            sorted_protein_sequences = []

        self.sorted_protein_sequences = sorted_protein_sequences

        if filter_values is None:
            self.filter_values = OrderedDict()
        else:
            self.filter_values = filter_values

        if phased_variants_in_supporting_reads is None:
            self.phased_variants_in_supporting_reads = set()
        else:
            self.phased_variants_in_supporting_reads = \
                phased_variants_in_supporting_reads

        if phased_variants_in_protein_sequence is None:
            self.phased_variants_in_protein_sequence = set()
        else:
            self.phased_variants_in_protein_sequence = \
                phased_variants_in_protein_sequence

    @property
    def fields(self):
        """
        List of field names used to construct an IsovarResult instance.
        """
        return [
            "variant",
            "predicted_effect",
            "read_evidence",
            "sorted_protein_sequences",
            "filter_values",
            "phased_variants_in_supporting_reads",
            "phased_variants_in_protein_sequence",
        ]

    def __str__(self):
        field_strings = ["%s=%s" % (k, v) for (k, v) in self.to_dict().items()]
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(field_strings)
        )

    def __eq__(self, other):
        for field_name in self.fields:
            my_value = getattr(self, field_name)
            other_value = getattr(other, field_name)
            if my_value is None and other_value is not None:
                return False
            elif other_value is None and my_value is not None:
                return False
            elif my_value != other_value:
                return False
        return True

    def __repr__(self):
        return str(self)

    def to_dict(self):
        """
        Dictionary representation of fields used to construct this IsovarResult

        Returns dict
        """
        return OrderedDict([
            (k, getattr(self, k))
            for k in self.fields
        ])

    def clone_with_updates(self, **kwargs):
        """
        Create a copy of this IsovarResult object including any new
        parameters in `kwargs`.

        Returns IsovarResult
        """
        for (k, v) in self.to_dict().items():
            if k not in kwargs:
                kwargs[k] = v
        return IsovarResult(**kwargs)

    def clone(self):
        """
        Make a copy of this IsovarResult without updating any fields.

        Returns IsovarResult
        """
        return self.clone_with_updates()

    def to_record(self):
        """
        Create an OrderedDict of essential information from
        this IsovarResult to be used for building a DataFrame across
        variants.

        Returns OrderedDict
        """
        d = OrderedDict([
            ("variant", self.variant.short_description),
            ("overlapping_gene_names",
                ";".join(self.overlapping_gene_names(only_coding=False))),
            ("overlapping_gene_ids",
                ";".join(self.overlapping_gene_ids(only_coding=False))),
            ("overlapping_coding_gene_names",
             ";".join(self.overlapping_gene_names(only_coding=True))),
            ("overlapping_coding_gene_ids",
             ";".join(self.overlapping_gene_ids(only_coding=True))),

        ])

        # get all quantitative fields from this object
        for key in dir(self):
            if key.startswith("num_") or key.startswith("fraction_") or key.startswith("ratio_"):
                d[key] = getattr(self, key)

        # get all boolean properties that start with "has_"
        for key in dir(self):
            if key.startswith("has_"):
                d[key] = getattr(self, key)

        ########################################################################
        # predicted protein changes without looking at RNA reads
        ########################################################################
        effect = self.predicted_effect

        d["predicted_effect"] = effect.short_description
        d["predicted_effect_class"] = effect.__class__.__name__
        d["predicted_effect_modifies_protein_sequence"] = \
            self.predicted_effect_modifies_protein_sequence

        # list of field names on varcode effect properties
        effect_properties = [
            "gene_name",
            "gene_id",
            "transcript_id",
            "transcript_name",
            "modifies_protein_sequence",
            "original_protein_sequence",
            "aa_mutation_start_offset",
            "aa_mutation_end_offset",
            "mutant_protein_sequence"
        ]
        for field_name in effect_properties:
            # store effect fields with prefix 'predicted_effect_' and use
            # getattr in case the field is not available for all effects
            d["predicted_effect_%s" % field_name] = getattr(
                effect,
                field_name,
                None)

        ########################################################################
        # get the top protein sequence, if one exists
        ########################################################################
        protein_sequence = self.top_protein_sequence

        # list of names we want to use in the result dictionary,
        # paired with names of fields on ProteinSequence
        protein_sequence_properties = [
            ("protein_sequence", "amino_acids"),
            ("protein_sequence_ends_with_stop_codon", "ends_with_stop_codon"),
            ("protein_sequence_gene_names", "gene_names"),
            ("protein_sequence_gene_ids", "gene_ids"),
            ("protein_sequence_transcript_names", "transcript_names"),
            ("protein_sequence_transcript_ids", "transcript_ids"),
        ]
        for (name, protein_sequence_field) in protein_sequence_properties:
            value = getattr(protein_sequence, protein_sequence_field, None)
            if isinstance(value, (list, set, tuple)):
                value = ";".join(value)
            d[name] = value
        d["protein_sequence_mutation_start"] = self.protein_sequence_mutation_start
        d["protein_sequence_mutation_end"] = self.protein_sequence_mutation_end
        d["num_mutant_amino_acids_in_protein_sequence"] = \
            self.num_mutant_amino_acids_in_protein_sequence

        d["trimmed_predicted_mutant_protein_sequence"] = self.trimmed_predicted_mutant_protein_sequence
        d["trimmed_reference_protein_sequence"] = self.trimmed_reference_protein_sequence
        d["protein_sequence_contains_mutation"] = self.protein_sequence_contains_mutation
        d["protein_sequence_matches_predicted_mutation_effect"] = self.protein_sequence_matches_predicted_mutation_effect

        ########################################################################
        # filters
        ########################################################################
        for filter_name, filter_value in self.filter_values.items():
            d["filter:%s" % filter_name] = filter_value
        d["passes_all_filters"] = self.passes_all_filters
        return d


    @cached_property
    def passes_all_filters(self):
        """
        Does this IsovarResult have True for all the filter values in
        self.filter_values?
        """
        if len(self.filter_values) == 0:
            return True
        else:
            return all(list(self.filter_values.values()))

    @cached_property
    def top_protein_sequence(self):
        """
        If any protein sequences were assembled for this variant then
        return the best according to coverage, number of mismatches
        relative to the reference, number of reference transcripts
        which match sequence before the variant and protein
        sequence length.

        Returns ProteinSequence or None
        """
        if len(self.sorted_protein_sequences) > 0:
            return self.sorted_protein_sequences[0]
        else:
            return None

    @cached_property
    def protein_sequence_mutation_start(self):
        """
        Interbase start coordinate for mutated amino acids in top protein
        sequence.

        Returns
        -------
        int or None
        """
        if self.has_mutant_protein_sequence_from_rna:
            return self.top_protein_sequence.mutation_start
        else:
            return None

    @cached_property
    def protein_sequence_mutation_end(self):
        """
        Interbase end coordinate for mutated amino acids in top protein
        sequence.

        Returns
        -------
        int or None
        """
        if self.has_mutant_protein_sequence_from_rna:
            return self.top_protein_sequence.mutation_end
        else:
            return None

    @cached_property
    def predicted_effect_modifies_protein_sequence(self):
        """
        Does the predicted effect change the protein sequence?

        Returns bool
        """
        if self.predicted_effect is None:
            return None
        return self.predicted_effect.modifies_protein_sequence

    @cached_property
    def num_cdna_mismatches_in_top_protein_sequence(self):
        """
        How many nucleotide positions aside from the variant location don't
        match the reference transcript sequence in the cDNA sequence used
        to translate the top protein sequence. When multiple cDNA sequences
        were used then take the min across them.

        Returns int
        """
        if self.top_protein_sequence is None:
            return None
        return self.top_protein_sequence.num_mismatches

    @cached_property
    def num_cdna_mismatches_before_variant_in_top_protein_sequence(self):
        """
        How many nucleotide positions before the variant location don't
        match the reference transcript sequence in the cDNA sequence used
        to translate the top protein sequence. When multiple cDNA sequences
        were used then take the min across them.

        Returns int
        """
        if self.top_protein_sequence is None:
            return None
        return self.top_protein_sequence.num_mismatches_before_variant

    @cached_property
    def num_cdna_mismatches_after_variant_in_top_protein_sequence(self):
        """
        How many nucleotide positions after the variant location don't
        match the reference transcript sequence in the cDNA sequence used
        to translate the top protein sequence. When multiple cDNA sequences
        were used then take the min across them.

        Returns int
        """
        if self.top_protein_sequence is None:
            return None
        return self.top_protein_sequence.num_mismatches_after_variant

    @cached_property
    def num_reads_supporting_top_protein_sequence(self):
        """
        How many reads support the assembled protein sequence?

        Returns int
        """
        if self.top_protein_sequence is None:
            return 0
        return self.top_protein_sequence.num_supporting_reads

    @cached_property
    def num_fragments_supporting_top_protein_sequence(self):
        """
        How many fragments (unique read names) support the assembled
        protein sequence?

        Returns int
        """
        if self.top_protein_sequence is None:
            return 0
        return self.top_protein_sequence.num_supporting_fragments

    @cached_property
    def num_translations_from_top_protein_sequence(self):
        """
        How many distinct translations were used to create the top
        protein sequence? These can arise either from different cDNA sequences
        or different reading frames.

        Returns int
        """
        if self.top_protein_sequence is None:
            return 0
        return len(self.top_protein_sequence.translations)

    @cached_property
    def num_cdna_sequences_from_top_protein_sequence(self):
        """
        How many distinct cDNA sequences were used to create the top
        protein sequence?

        Returns int
        """
        if self.top_protein_sequence is None:
            return 0
        return self.top_protein_sequence.num_cdna_sequences

    @cached_property
    def trimmed_predicted_mutant_protein_sequence(self):
        """
        Trim the predicted mutant protein sequence from Varcode
        to match the length of the protein subsequence assembled from RNA.

        Returns str or None
        """
        p = self.top_protein_sequence
        e = self.predicted_effect

        if e is None or p is None:
            return None
        if e.mutant_protein_sequence is None:
            return None
        if e.aa_mutation_start_offset is None:
            return None
        n_before_mutation = p.mutation_start
        n_after_mutation = len(p.amino_acids) - p.mutation_end
        return e.mutant_protein_sequence[
             e.aa_mutation_start_offset - n_before_mutation:
             e.aa_mutation_start_offset + len(e.aa_alt) + n_after_mutation]

    @cached_property
    def has_mutant_protein_sequence_from_rna(self):
        """
        Does this IsovarResult have an associated protein sequence
        determined from RNA reads?

        Returns bool
        """
        return self.top_protein_sequence is not None

    @cached_property
    def has_reference_protein_sequence_from_predicted_effect(self):
        """
        Does the variant for this IsovarResult affect a protein
        whose reference sequence can be determined?

        Returns bool
        """
        if self.predicted_effect is None:
            return False
        return self.predicted_effect.original_protein_sequence is not None

    @cached_property
    def has_mutant_protein_sequence_from_predicted_effect(self):
        """
        Does the variant for this IsovarResult have a predicted
        mutant protein sequence?

        Returns bool
        """
        if self.predicted_effect is None:
            return False
        return self.predicted_effect.mutant_protein_sequence is not None

    @cached_property
    def trimmed_reference_protein_sequence(self):
        """
        Trim the reference protein sequence from the top Varcode effect
        to match the length of the protein sequence assembled from RNA.

        Returns str
        """
        p = self.top_protein_sequence
        e = self.predicted_effect
        if e is None or p is None:
            return None
        if e.original_protein_sequence is None:
            return None
        if e.aa_mutation_start_offset is None:
            return None
        n_before_mutation = p.mutation_start
        n_total = len(p.amino_acids)
        start_index_in_original_protein = (
               e.aa_mutation_start_offset - n_before_mutation)
        return e.original_protein_sequence[
               start_index_in_original_protein:
               start_index_in_original_protein + n_total]

    @cached_property
    def num_amino_acid_mismatches_from_predicted_effect(self):
        """
        Compute the number of mismatches between the mutant protein sequence
        predicted by Varcode and the best supported sequence translated
        from assembled RNA reads by Isovar. We're not allowing any
        insertions or deletions in the middle of the sequences but do
        allow a shorter sequence to start anywhere within a longer one.

        Returns int
        """
        predicted_sequence = self.trimmed_predicted_mutant_protein_sequence
        if predicted_sequence is None:
            return None

        protein_sequence_object = self.top_protein_sequence
        if protein_sequence_object is None:
            return None

        assembled_protein_sequence = protein_sequence_object.amino_acids

        return alignment_score(
            assembled_protein_sequence,
            predicted_sequence)

    @cached_property
    def protein_sequence_matches_predicted_mutation_effect(self):
        """
        Does the top protein sequence translated from RNA reads
        match the predicted protein change determined by Varcode?

        Returns bool
        """
        n_mismatches = self.num_amino_acid_mismatches_from_predicted_effect
        if n_mismatches is None:
            return None
        return n_mismatches == 0


    @cached_property
    def num_amino_acid_mismatches_from_reference(self):
        """
        Compute the number of mismatches between the original protein sequence
        and the best supported sequence translated from assembled RNA reads by
        Isovar. We're not allowing any insertions or deletions in the middle of
        the sequences but do allow a shorter sequence to start anywhere
        within a longer one.

        Returns int
        """
        original_sequence = self.trimmed_reference_protein_sequence
        if original_sequence is None:
            return None

        protein_sequence_object = self.top_protein_sequence
        if protein_sequence_object is None:
            return None
        assembled_protein_sequence = protein_sequence_object.amino_acids

        return alignment_score(
            assembled_protein_sequence,
            original_sequence)

    @cached_property
    def protein_sequence_contains_mutation(self):
        """
        Does the protein sequence assembled from RNA contain a mutation?

        Returns bool
        """
        if not self.has_mutant_protein_sequence_from_rna:
            return None

        # if the sequence is the same as the reference then the genomic
        # variant must have been silent, or maybe was made silent by adjacent
        # phased variants.
        if self.protein_sequence_matches_reference:
            return False

        start_idx  = self.protein_sequence_mutation_start
        stop_idx = self.protein_sequence_mutation_end

        if start_idx is None or stop_idx is None:
            return False
        return start_idx != stop_idx or self.variant.is_deletion

    @cached_property
    def num_mutant_amino_acids_in_protein_sequence(self):
        """
        Number of amino acids modified by mutation in the top protein sequence
        from RNA.

        Returns
        -------
        int or None
        """
        if self.top_protein_sequence is None:
            return None
        return self.top_protein_sequence.num_mutant_amino_acids

    @cached_property
    def protein_sequence_matches_reference(self):
        """
        Does the top protein sequence translated from RNA reads
        match the reference protein sequence?

        Returns bool
        """
        n_mismatches = self.num_amino_acid_mismatches_from_reference
        if n_mismatches is None:
            return None
        return n_mismatches == 0

    @cached_property
    def protein_sequence_different_from_reference(self):
        """
        Does the top protein sequence translated from RNA reads
        differ from the reference?

        Returns bool
        """
        if self.protein_sequence_matches_reference is None:
            return None
        return not self.protein_sequence_matches_reference

    @cached_property
    def num_protein_sequences(self):
        """
        Number of distinct protein sequences which were translated from
        assembled RNA reads.

        Returns int
        """
        return len(self.sorted_protein_sequences)

    def transcripts_from_protein_sequences(self, max_num_protein_sequences=None):
        """
        Ensembl transcript IDs of all transcripts which support the reading
        frame used by protein sequences in this IsovarResult.

        Parameters
        ----------
        max_num_protein_sequences : int or None
            If supplied then only consider the top protein sequences up to
            this number.

        Returns list of pyensembl.Transcript
        """
        transcript_set = set([])
        for p in self.sorted_protein_sequences[:max_num_protein_sequences]:
            transcript_set.update(p.transcripts)
        return sorted(transcript_set)

    def transcript_ids_from_protein_sequences(self, max_num_protein_sequences=None):
        """
        Ensembl transcripts IDs which support the reading frame used by protein
        sequences in this IsovarResult.

        Parameters
        ----------
        max_num_protein_sequences : int or None
            If supplied then only consider the top protein sequences up to
            this number.

        Returns list of str
        """
        return sorted({t.id for t in self.transcripts_from_protein_sequences(
            max_num_protein_sequences=max_num_protein_sequences)})

    @cached_property
    def num_transcripts_from_protein_sequences(self):
        """
        Number of genes used by any translated protein sequence associated
        with this IsovarResult.

        Returns int
        """
        return len(self.transcript_ids_from_protein_sequences(
            max_num_protein_sequences=None))

    @cached_property
    def num_transcripts_from_top_protein_sequence(self):
        """
        Number of genes used by any translated protein sequence associated
        with this IsovarResult.

        Returns int
        """
        return len(self.transcript_ids_from_protein_sequences(
            max_num_protein_sequences=1))

    def genes_from_protein_sequences(self, max_num_protein_sequences=None):
        """
        Ensembl genes which support the reading frame used by protein
        sequences in this IsovarResult.

        Parameters
        ----------
        max_num_protein_sequences : int or None
            If supplied then only consider the top protein sequences up to
            this number.

        Returns list of pyensembl.Gene
        """
        transcripts = self.transcripts_from_protein_sequences(
            max_num_protein_sequences=max_num_protein_sequences)
        genes = [t.gene for t in transcripts]
        return sorted(genes)

    def gene_ids_from_protein_sequences(self, max_num_protein_sequences=None):
        """
        Ensembl genes IDs which support the reading frame used by protein
        sequences in this IsovarResult.

        Parameters
        ----------
        max_num_protein_sequences : int or None
            If supplied then only consider the top protein sequences up to
            this number.

        Returns list of str
        """
        return sorted({
            g.id
            for g
            in
            self.genes_from_protein_sequences(
                max_num_protein_sequences=max_num_protein_sequences)
        })

    @cached_property
    def num_genes_from_protein_sequences(self):
        """
        Number of genes used by any translated protein sequence associated
        with this IsovarResult.

        Returns int
        """
        return len(self.gene_ids_from_protein_sequences(
            max_num_protein_sequences=None))

    @cached_property
    def num_genes_from_top_protein_sequence(self):
        """
        Number of genes used by any translated protein sequence associated
        with this IsovarResult.

        Returns int
        """
        return len(self.gene_ids_from_protein_sequences(
            max_num_protein_sequences=1))

    def overlapping_transcripts(self, only_coding=True):
        """
        Transcripts which this variant overlaps.

        Parameters
        ----------
        only_coding : bool
            Only return transcripts which are annotated as coding for a
            protein (default=True)

        Returns set of pyensembl.Transcript objects
        """
        return {
            t
            for t in self.variant.transcripts
            if not only_coding or t.is_protein_coding
        }

    def overlapping_transcript_ids(self, only_coding=True):
        """
        Transcript IDs which this variant overlaps.

        Parameters
        ----------
        only_coding : bool
            Only return transcripts which are annotated as coding for a
            protein (default=True)
        Returns set of str
        """
        return {
            t.id
            for t in self.variant.transcripts
            if not only_coding or t.is_protein_coding
        }

    @cached_property
    def num_overlapping_transcripts(self):
        """
        Number of transcripts overlapped by the variant

        Returns int
        """
        return len(self.overlapping_transcript_ids(only_coding=False))

    @cached_property
    def num_overlapping_coding_transcripts(self):
        """
        Number of coding transcripts overlapped by the variant

        Returns int
        """
        return len(self.overlapping_transcript_ids(only_coding=True))

    def overlapping_genes(self, only_coding=True):
        """
        Genes which this variant overlaps.

        Parameters
        ----------
        only_coding : bool
            Only return genes which are annotated as coding for a
            protein (default=True)

        Returns list of pyensembl.Gene objects
        """
        return sorted({
            g
            for g in self.variant.genes
            if not only_coding or g.is_protein_coding
        })

    def overlapping_gene_names(self, only_coding=True):
        """
        Names of genes which this variant overlaps.

        Parameters
        ----------
        only_coding : bool
           Only return genes which are annotated as coding for a
           protein (default=True)

        Returns list of str
        """
        return [
            g.name for g in self.overlapping_genes(only_coding=only_coding)
        ]

    def overlapping_gene_ids(self, only_coding=True):
        """
        Gene IDs which this variant overlaps.

        Parameters
        ----------
        only_coding : bool
            Only return genes which are annotated as coding for a
            protein (default=True)

        Returns set of str
        """
        return {
            g.id
            for g in self.variant.genes
            if not only_coding or g.is_protein_coding
        }

    @cached_property
    def num_overlapping_genes(self):
        """
        Number of genes overlapped by the variant

        Returns int
        """
        return len(self.overlapping_gene_ids(only_coding=False))

    @cached_property
    def num_overlapping_coding_genes(self):
        """
        Number of coding genes overlapped by the variant

        Returns int
        """
        return len(self.overlapping_gene_ids(only_coding=True))

    @cached_property
    def ref_reads(self):
        """
        AlleleRead objects at this locus which support the reference allele
        """
        return self.read_evidence.ref_reads

    @cached_property
    def alt_reads(self):
        """
        AlleleRead objects at this locus which support the mutant allele
        """
        return self.read_evidence.alt_reads

    @cached_property
    def other_reads(self):
        """
        AlleleRead objects at this locus which support some allele other than
        either the reference or alternate.
        """
        return self.read_evidence.other_reads

    @cached_property
    def ref_read_names(self):
        """
        Names of reference reads at this locus.
        """
        return {r.name for r in self.ref_reads}

    @cached_property
    def alt_read_names(self):
        """
        Names of alt reads at this locus.
        """
        return {r.name for r in self.alt_reads}

    @cached_property
    def ref_and_alt_read_names(self):
        """
        Names of reads which support either the ref or alt alleles.
        """
        return self.ref_read_names.union(self.alt_read_names)

    @cached_property
    def other_read_names(self):
        """
        Names of other (non-alt, non-ref) reads at this locus.
        """
        return {r.name for r in self.other_reads}

    @cached_property
    def all_read_names(self):
        """
        Names of all reads at this locus.
        """
        return self.ref_read_names.union(self.alt_read_names).union(self.other_read_names)

    @cached_property
    def num_total_reads(self):
        """
        Total number of reads at this locus, regardless of allele.
        """
        return self.num_ref_reads + self.num_alt_reads + self.num_other_reads

    @cached_property
    def num_total_fragments(self):
        """
        Total number of distinct fragments at this locus, which also corresponds
        to the total number of read names.
        """
        return len(self.all_read_names)

    @cached_property
    def num_ref_reads(self):
        """
        Number of reads which support the reference allele.
        """
        return len(self.ref_reads)

    @cached_property
    def num_ref_fragments(self):
        """
        Number of distinct fragments which support the reference allele.
        """
        return len(self.ref_read_names)

    @cached_property
    def num_alt_reads(self):
        """
        Number of reads which support the alt allele.
        """
        return len(self.alt_reads)

    @cached_property
    def num_alt_fragments(self):
        """
        Number of distinct fragments which support the alt allele.
        """
        return len(self.alt_read_names)

    @cached_property
    def num_other_reads(self):
        """
        Number of reads which support neither the reference nor alt alleles.
        """
        return len(self.other_reads)

    @cached_property
    def num_other_fragments(self):
        """
        Number of distinct fragments which support neither the reference nor
        alt alleles.
        """
        return len(self.other_read_names)

    @cached_property
    def fraction_ref_reads(self):
        """
        Allelic fraction of the reference allele among all reads at this site.
        """
        return safediv(self.num_ref_reads, self.num_total_reads)

    @cached_property
    def fraction_ref_fragments(self):
        """
        Allelic fraction of the reference allele among all fragments at this site.
        """
        return safediv(self.num_ref_fragments, self.num_total_fragments)

    @cached_property
    def fraction_alt_reads(self):
        """
        Allelic fraction of the variant allele among all reads at this site.
        """
        return safediv(self.num_alt_reads, self.num_total_reads)

    @cached_property
    def fraction_alt_fragments(self):
        """
        Allelic fraction of the variant allele among all fragments at this site.
        """
        return safediv(self.num_alt_fragments, self.num_total_fragments)

    @cached_property
    def fraction_other_reads(self):
        """
        Allelic fraction of the "other" (non-ref, non-alt) alleles among all
        reads at this site.
        """
        return safediv(self.num_other_reads, self.num_total_reads)

    @cached_property
    def fraction_other_fragments(self):
        """
        Allelic fraction of the "other" (non-ref, non-alt) alleles among all
        fragments at this site.
        """
        return safediv(self.num_other_fragments, self.num_total_fragments)

    @cached_property
    def ratio_other_to_ref_reads(self):
        """
        Ratio of the number of reads which support alleles which are neither
        ref/alt to the number of ref reads.
        """
        return safediv(self.num_other_reads, self.num_ref_reads)

    @cached_property
    def ratio_other_to_ref_fragments(self):
        """
        Ratio of the number of fragments which support alleles which are neither
        ref/alt to the number of ref fragments.
        """
        return safediv(self.num_other_fragments, self.num_ref_fragments)

    @cached_property
    def ratio_other_to_alt_reads(self):
        """
        Ratio of the number of reads which support alleles which are neither
        ref/alt to the number of alt reads.
        """
        return safediv(self.num_other_reads, self.num_alt_reads)

    @cached_property
    def ratio_other_to_alt_fragments(self):
        """
        Ratio of the number of fragments which support alleles which are neither
        ref/alt to the number of alt fragments.
        """
        return safediv(self.num_other_fragments, self.num_alt_fragments)

    @cached_property
    def ratio_ref_to_other_reads(self):
        """
        Ratio of the number of reference reads to non-ref/non-alt reads
        """
        return safediv(self.num_ref_reads, self.num_other_reads)

    @cached_property
    def ratio_ref_to_other_fragments(self):
        """
        Ratio of the number of reference fragments to non-ref/non-alt fragments
        """
        return safediv(self.num_ref_fragments, self.num_other_fragments)

    @cached_property
    def ratio_alt_to_other_reads(self):
        """
        Ratio of alt allele reads to non-ref/non-alt reads
        """
        return safediv(self.num_alt_reads, self.num_other_reads)

    @cached_property
    def ratio_alt_to_other_fragments(self):
        """
        Ratio of the number of fragments which support the alt allele
        to the number of non-alt/non-ref allele fragments.
        """
        return safediv(self.num_alt_fragments, self.num_other_fragments)


    @cached_property
    def num_phased_variants_from_reads(self):
        """
        Number of variants explicitly phased with the variant in this
        IsovarResult by shared RNA fragments.

        Returns int
        """
        return len(self.phased_variants_in_supporting_reads)

    @cached_property
    def num_phased_variants_in_protein_sequence(self):
        """
        Number of variants explicitly phased with the variant in this
        IsovarResult by shared RNA fragments which were used to create
        the top protein sequences for this variant and whatever it
        gets phased with.

        Returns int
        """
        return len(self.phased_variants_in_protein_sequence)


# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common command-line arguments for all Isovar commands which use RNA
"""

from __future__ import print_function, division, absolute_import

from pysam import AlignmentFile

from varcode.cli import make_variants_parser, variant_collection_from_args

from ..default_parameters import MIN_READ_MAPPING_QUALITY

from ..read_collector import ReadCollector
from ..dataframe_helpers import allele_reads_to_dataframe, read_evidence_generator_to_dataframe


def add_rna_args(
        parser,
        min_mapping_quality_default=MIN_READ_MAPPING_QUALITY):
    """
    Extends an ArgumentParser instance with the following commandline arguments:
        --bam
        --min-mapping-quality
        --use-duplicate-reads
        --drop-secondary-alignments
    """
    rna_group = parser.add_argument_group("RNA")
    rna_group.add_argument(
        "--bam",
        required=True,
        help="BAM file containing RNAseq reads")

    rna_group.add_argument(
        "--min-mapping-quality",
        type=int,
        default=min_mapping_quality_default,
        help="Minimum MAPQ value to allow for a read (default %(default)s)")

    rna_group.add_argument(
        "--use-duplicate-reads",
        default=False,
        action="store_true",
        help=(
            "By default, reads which have been marked as duplicates are excluded."
            "Use this option to include duplicate reads."))

    rna_group.add_argument(
        "--drop-secondary-alignments",
        default=False,
        action="store_true",
        help=(
            "By default, secondary alignments are included in reads, "
            "use this option to instead only use primary alignments."))

    rna_group.add_argument(
        "--num-rna-decompression-threads",
        type=int,
        help=(
            "Number of threads to use for decompression of BAM/CRAM files "
            "(default %(default)s)."),
        default=1)

    return rna_group


def make_rna_reads_arg_parser(**kwargs):
    """
    Parameters
    ----------
    **kwargs : dict
        Passed directly to argparse.ArgumentParser

    Creates argparse.ArgumentParser instance with all of the options
    needed to load a set of variants and their supporting RNA reads.
    """
    parser = make_variants_parser(**kwargs)
    add_rna_args(parser)
    return parser


def alignment_file_from_args(args):
    """
    Use parsed arguments to load a file of aligned RNA reads.
    """
    return AlignmentFile(
        args.bam,
        threads=args.num_rna_decompression_threads)


def read_collector_from_args(args):
    """
    Use parsed arguments to create a ReadCollector object
    """
    return ReadCollector(
        min_mapping_quality=args.min_mapping_quality,
        use_duplicate_reads=args.use_duplicate_reads,
        use_secondary_alignments=not args.drop_secondary_alignments)


def read_evidence_generator_from_args(args):
    """
    Creates a generator of (Variant, ReadEvidence) pairs from parsed
    arguments.
    """
    variants = variant_collection_from_args(args)
    samfile = alignment_file_from_args(args)
    read_creator = read_collector_from_args(args)
    return read_creator.read_evidence_generator(
        variants=variants,
        alignment_file=samfile)

def variant_reads_generator_from_args(args):
    """
    Creates a generator of (Variant, list of AlleleRead) from parsed
    arguments, where all AlleleRead objects must have alleles matching
    the variant.
    """
    for variant, read_evidence in read_evidence_generator_from_args(args):
        yield variant, read_evidence.alt_reads


def read_evidence_dataframe_from_args(args):
    """
    Collect ReadEvidence for each variant and turn them into a DataFrame
    """
    return read_evidence_generator_to_dataframe(
        read_evidence_generator_from_args(args))


def variants_reads_dataframe_from_args(args):
    """
    Collect variant reads for each variant and turn them into a DataFrame
    """
    return allele_reads_to_dataframe(
        read_evidence_generator_from_args(args))



# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from six import string_types
from varcode import load_vcf
from pysam import AlignmentFile
from collections import OrderedDict

from .protein_sequence_creator import ProteinSequenceCreator
from .read_collector import ReadCollector
from .logging import get_logger
from .isovar_result import IsovarResult
from .default_parameters import (
    MIN_NUM_RNA_ALT_READS,
    MIN_NUM_RNA_ALT_FRAGMENTS,
    MIN_FRACTION_RNA_ALT_READS,
    MIN_FRACTION_RNA_ALT_FRAGMENTS,
    MAX_NUM_RNA_REF_READS,
    MAX_NUM_RNA_REF_FRAGMENTS,
    MAX_FRACTION_RNA_REF_READS,
    MAX_FRACTION_RNA_REF_FRAGMENTS,
    MAX_NUM_RNA_OTHER_READS,
    MAX_NUM_RNA_OTHER_FRAGMENTS,
    MAX_FRACTION_RNA_OTHER_READS,
    MAX_FRACTION_RNA_OTHER_FRAGMENTS,
    MIN_RATIO_RNA_ALT_TO_OTHER_FRAGMENTS,
    MIN_SHARED_FRAGMENTS_FOR_PHASING
)
from .effect_prediction import top_varcode_effect
from .filtering import apply_filters
from .phasing import annotate_phased_variants

logger = get_logger(__name__)


DEFAULT_FILTER_THRESHOLDS =  OrderedDict([
    # alt allele
    ("min_num_alt_reads", MIN_NUM_RNA_ALT_READS),
    ("min_num_alt_fragments", MIN_NUM_RNA_ALT_FRAGMENTS),
    ("min_fraction_alt_reads", MIN_FRACTION_RNA_ALT_READS),
    ("min_fraction_alt_fragments", MIN_FRACTION_RNA_ALT_FRAGMENTS),

    # ref allele coverage and VAF
    ("max_num_ref_reads", MAX_NUM_RNA_REF_READS),
    ("max_num_ref_fragments", MAX_NUM_RNA_REF_FRAGMENTS),
    ("max_fraction_ref_reads", MAX_FRACTION_RNA_REF_READS),
    ("max_fraction_ref_fragments", MAX_FRACTION_RNA_REF_FRAGMENTS),

    # other alleles
    ("max_num_other_reads", MAX_NUM_RNA_OTHER_READS),
    ("max_num_other_fragments", MAX_NUM_RNA_OTHER_FRAGMENTS),
    ("max_fraction_other_reads", MAX_FRACTION_RNA_OTHER_READS),
    ("max_fraction_other_fragments", MAX_FRACTION_RNA_OTHER_FRAGMENTS),

    # misc. filters
    ("min_ratio_alt_to_other_fragments", MIN_RATIO_RNA_ALT_TO_OTHER_FRAGMENTS)
])

DEFAULT_FILTER_FLAGS = [
    "predicted_effect_modifies_protein_sequence",
    "has_mutant_protein_sequence_from_rna",
    "protein_sequence_contains_mutation",
]

def run_isovar(
        variants,
        alignment_file,
        transcript_id_whitelist=None,
        read_collector=None,
        protein_sequence_creator=None,
        filter_thresholds=DEFAULT_FILTER_THRESHOLDS,
        filter_flags=DEFAULT_FILTER_FLAGS,
        min_shared_fragments_for_phasing=MIN_SHARED_FRAGMENTS_FOR_PHASING,
        decompression_threads=1):
    """
    This is the main entrypoint into the Isovar library, which collects
    RNA reads supporting variants and translates their coding sequence
    into amino acid sequences. Collects both the read evidence and
    protein sequences into IsovarResult objects. The values of any filters
    which are supplied in the filter_thresholds argument are attached to
    each IsovarResult's filter_values_dict field.

    Parameters
    ----------
    variants : varcode.VariantCollection
        Somatic variants

    alignment_file : pysam.AlignmentFile
        Aligned tumor RNA reads

    transcript_id_whitelist : set of str or None
        Which transcripts should be considered when predicting DNA-only
        coding effects of mutations and also when trying to establish a
        reading frame for identified cDNA sequences.

    read_collector : ReadCollector or None
        Object used to collect ReadEvidence for each variant, created
        with default settings if not supplied.

    protein_sequence_creator : ProteinSequenceCreator or None
        Object used to turn (Variant, ReadEvidence) into one or more
        ProteinSequence objects. Created with default settings if not
        supplied.

    filter_thresholds : dict or OrderedDict
        Dictionary whose entries have names like "min_num_alt_reads"
        mapping to a numerical threshold value. In general, the keys
        must start with either "min_" or "max_" followed by a property
        of the IsovarResult class.

    filter_flags : list of str
        List of boolean fields of IsovarResult used for filtering,
        they can also be negated by prepending "not_",
        such as "not_has_protein_sequence".

    decompress_threads : int
        Number of threads used by htslib to decompress BAM/CRAM
        files.

    Generator of IsovarResult objects, one for each variant. The
    `protein_sequences` field of the IsovarVar result will be empty
    if no sequences could be determined.
    """
    if isinstance(variants, string_types):
        variants = load_vcf(variants)

    if isinstance(alignment_file, string_types):
        alignment_file = AlignmentFile(
            alignment_file,
            threads=decompression_threads)

    if read_collector is None:
        read_collector = ReadCollector()

    if protein_sequence_creator is None:
        protein_sequence_creator = ProteinSequenceCreator()

    # create generator which returns (Variant, ReadEvidence) pairs
    read_evidence_gen = \
        read_collector.read_evidence_generator(
           variants=variants,
           alignment_file=alignment_file)

    results = []
    for variant, read_evidence in read_evidence_gen:
        # generate protein sequences by assembling variant reads
        protein_sequences = \
            protein_sequence_creator.sorted_protein_sequences_for_variant(
                variant=variant,
                read_evidence=read_evidence,
                transcript_id_whitelist=transcript_id_whitelist)
        predicted_effect = top_varcode_effect(
            variant=variant,
            transcript_id_whitelist=transcript_id_whitelist)
        isovar_result = IsovarResult(
            variant=variant,
            predicted_effect=predicted_effect,
            read_evidence=read_evidence,
            sorted_protein_sequences=protein_sequences)
        isovar_result = apply_filters(
            isovar_result,
            filter_thresholds=filter_thresholds,
            filter_flags=filter_flags)
        results.append(isovar_result)
    results = annotate_phased_variants(
        results,
        min_shared_fragments_for_phasing)
    return results



from __future__ import print_function, division, absolute_import

from pyensembl import ensembl_grch38

from varcode import Variant
from nose.tools import eq_

from testing_helpers import load_bam

from isovar import ReadCollector


def test_partition_variant_reads_snv():
    alignment_file = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 65857041
    ref = "G"
    alt = "C"
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref,
        alt=alt,
        ensembl=ensembl_grch38)
    read_collector = ReadCollector()
    read_evidence = read_collector.read_evidence_for_variant(
        alignment_file=alignment_file,
        variant=variant)
    alt_reads = read_evidence.alt_reads
    assert len(alt_reads) > 1
    for variant_read in alt_reads:
        eq_(variant_read.allele, alt)


def test_partition_variant_reads_deletion():
    alignment_file = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 70091490
    ref = "TTGTAGATGCTGCCTCTCC"
    alt = ""
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref,
        alt=alt,
        ensembl=ensembl_grch38)
    read_collector = ReadCollector()
    read_evidence = read_collector.read_evidence_for_variant(
        alignment_file=alignment_file,
        variant=variant)
    assert len(read_evidence.alt_reads) > 1
    for variant_read in read_evidence.alt_reads:
        eq_(variant_read.allele, alt)

if __name__ == "__main__":
    test_partition_variant_reads_snv()
    test_partition_variant_reads_deletion()



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .logging import get_logger
from .variant_orf import VariantORF

logger = get_logger(__name__)


def match_variant_sequence_to_reference_context(
        variant_sequence,
        reference_context,
        min_transcript_prefix_length,
        max_transcript_mismatches,
        count_mismatches_after_variant=False,
        max_trimming_attempts=2):
    """
    Iteratively trim low-coverage subsequences of a variant sequence
    until it either matches the given reference context or there
    are too few nucleotides left in the variant sequence.

    Parameters
    ----------
    variant_sequence : VariantSequence
        Assembled sequence from RNA reads, will need to be to be reverse
        complemented if matching against a reference transcript on the
        negative strand.

    reference_context : ReferenceContext
        Sequence of reference transcript before the variant and associated
        metadata.

    min_transcript_prefix_length : int
        Minimum number of nucleotides we try to match against a reference
        transcript.

    max_transcript_mismatches : int
        Maximum number of nucleotide differences between reference transcript
        sequence and the variant sequence.

    count_mismatches_after_variant : bool
        Set to true if the number of mismatches after the variant locus should
        count toward the total max_transcript_mismatches, which by default
        only counts mismatches before the variant locus.

    max_trimming_attempts : int
        How many times do we try trimming the VariantSequence to higher
        levels of coverage before giving up?

    Returns VariantORF or None
    """
    # if we can't get the variant sequence to match this reference
    # context then keep trimming it by coverage until either
    for i in range(max_trimming_attempts + 1):
        # check the reverse-complemented prefix if the reference context is
        # on the negative strand since variant sequence is aligned to
        # genomic DNA (positive strand)
        variant_sequence_too_short = (
            (reference_context.strand == "+" and
                len(variant_sequence.prefix) < min_transcript_prefix_length) or
            (reference_context.strand == "-" and
                len(variant_sequence.suffix) < min_transcript_prefix_length)
        )
        if variant_sequence_too_short:
            logger.info(
                "Prefix of variant sequence %s shorter than min allowed %d (iter=%d)",
                variant_sequence,
                min_transcript_prefix_length,
                i + 1)
            return None

        variant_orf = \
            VariantORF.from_variant_sequence_and_reference_context(
                variant_sequence=variant_sequence,
                reference_context=reference_context)

        if variant_orf is None:
            return None

        n_mismatch_before_variant = (
            variant_orf.num_mismatches_before_variant)
        n_mismatch_after_variant = (
            variant_orf.num_mismatches_after_variant)

        logger.info("Iter #%d/%d: %s (len=%d)" % (
            i + 1,
            max_trimming_attempts + 1,
            variant_orf,
            len(variant_orf.cdna_sequence)))

        total_mismatches = n_mismatch_before_variant
        if count_mismatches_after_variant:
            total_mismatches += n_mismatch_after_variant
        if total_mismatches <= max_transcript_mismatches:
            # if we got a variant sequence + reading frame with sufficiently
            # few mismatches then call it a day
            return variant_orf

        logger.info(
            ("Too many mismatches (%d) between variant sequence %s and "
             "reference context %s (attempt=%d/%d)"),
            n_mismatch_before_variant,
            variant_sequence,
            reference_context,
            i + 1,
            max_trimming_attempts + 1)
        # if portions of the sequence are supported by only 1 read
        # then try trimming to 2 to see if the better supported
        # subsequence can be better matched against the reference
        current_min_coverage = variant_sequence.min_coverage()
        logger.info(
            "Trimming to subsequence covered by at least %d reads",
            current_min_coverage + 1)
        variant_sequence = variant_sequence.trim_by_coverage(
            current_min_coverage + 1)
    return None



# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Primary Isovar command, used to collect information about variants,
the RNA reads which overlap and protein sequences which can be constructed
from reads that support the variant.
"""
from __future__ import print_function, division, absolute_import

import sys



from ..logging import get_logger
from ..dataframe_helpers import isovar_results_to_dataframe

from .main_args import run_isovar_from_parsed_args, make_isovar_arg_parser

from .output_args import add_output_args, write_dataframe

logger = get_logger(__name__)

def run(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = make_isovar_arg_parser()
    parser = add_output_args(
        parser,
        filename="isovar-results.csv")
    args = parser.parse_args(args)
    logger.info(args)
    isovar_results = run_isovar_from_parsed_args(args)
    df = isovar_results_to_dataframe(isovar_results)
    logger.info(df)
    write_dataframe(df, args)



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from collections import defaultdict


from .effect_prediction import reference_coding_transcripts_for_variant
from .reference_context import ReferenceContext
from .reference_coding_sequence_key import ReferenceCodingSequenceKey


def reference_contexts_for_variant(
        variant,
        context_size,
        transcript_id_whitelist=None):
    """
    variant : varcode.Variant

    context_size : int
        Max of nucleotides to include to the left and right of the variant
        in the context sequence.

    transcript_id_whitelist : set, optional
        If given, then only consider transcripts whose IDs are in this set.

    Returns list of ReferenceContext objects, sorted by maximum length of
    coding sequence of any supporting transcripts.
    """
    overlapping_transcripts = reference_coding_transcripts_for_variant(
        variant=variant,
        transcript_id_whitelist=transcript_id_whitelist)

    # dictionary mapping SequenceKeyWithReadingFrame keys to list of
    # transcript objects
    sequence_groups = defaultdict(list)

    for transcript in overlapping_transcripts:
        reference_coding_sequence_key = \
            ReferenceCodingSequenceKey.from_variant_and_transcript(
                variant=variant,
                transcript=transcript,
                context_size=context_size)
        if reference_coding_sequence_key is not None:
            sequence_groups[reference_coding_sequence_key].append(transcript)

    reference_contexts = [
        ReferenceContext.from_reference_coding_sequence_key(
            key, variant, matching_transcripts)
        for (key, matching_transcripts) in sequence_groups.items()
    ]
    reference_contexts.sort(
        key=ReferenceContext.sort_key_decreasing_max_length_transcript_cds)
    return reference_contexts


def reference_contexts_generator(
        variants,
        context_size,
        transcript_id_whitelist=None):
    """
    Extract a set of reference contexts for each variant in the collection.

    Parameters
    ----------
    variants : varcode.VariantCollection

    context_size : int
        Max of nucleotides to include to the left and right of the variant
        in the context sequence.

    transcript_id_whitelist : set, optional
        If given, then only consider transcripts whose IDs are in this set.

    Generate a series of (Variant, [ReferenceContext]) pairs, where the
    to list of ReferenceContext objects for each variant is sorted by
    max coding sequence length of any transcript.
    """
    for variant in variants:
        reference_contexts = reference_contexts_for_variant(
            variant=variant,
            context_size=context_size,
            transcript_id_whitelist=transcript_id_whitelist)
        yield variant, reference_contexts


# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads overlapping a locus of interest split into prefix,
allele (ref, alt, or otherwise), and suffix portions
"""

from __future__ import print_function, division, absolute_import
import logging

from .string_helpers import convert_from_bytes_if_necessary, trim_N_nucleotides
from .value_object import ValueObject

logger = logging.getLogger(__name__)


class AlleleRead(ValueObject):
    """
    Extremely simplified representation of a read at a locus: just the allele
    at the locus and sequence before/after. We're ignoring the base qualities
    and any additional information about splicing, clipping or alignment.
    """
    __slots__ = ["prefix", "allele", "suffix", "name", "sequence"]

    def __init__(self, prefix, allele, suffix, name):
        self.prefix = prefix
        self.allele = allele
        self.suffix = suffix
        self.name = name
        self.sequence = prefix + allele + suffix

    def __len__(self):
        return len(self.sequence)

    @classmethod
    def from_locus_read(cls, locus_read):
        """
        Given a single LocusRead object, return either an AlleleRead or None

        Parameters
        ----------
        locus_read : LocusRead
            Read which overlaps a variant locus but doesn't necessarily contain the
            alternate nucleotides
        """
        sequence = locus_read.sequence
        read_name = locus_read.name

        reference_base0_start_inclusive = locus_read.reference_base0_start_inclusive
        reference_base0_end_exclusive = locus_read.reference_base0_end_exclusive

        read_base0_start_inclusive = locus_read.read_base0_start_inclusive
        read_base0_end_exclusive = locus_read.read_base0_end_exclusive

        if read_base0_start_inclusive is None or read_base0_end_exclusive is None:
            logger.debug(
                "Skipping read '%s' because required bases in reference interval %s:%s aren't mapped",
                read_name,
                reference_base0_start_inclusive,
                reference_base0_end_exclusive)
            return None

        reference_positions = locus_read.reference_positions

        n_ref_bases = reference_base0_end_exclusive - reference_base0_start_inclusive

        insertion = (n_ref_bases == 0)

        if insertion:
            # insertions require a sequence of non-aligned bases
            # followed by the subsequence reference position
            for read_index in range(read_base0_start_inclusive, read_base0_end_exclusive):
                # all the inserted nucleotides should *not* align to the reference
                if reference_positions[read_index] is not None:
                    logger.debug(
                        "Skipping read '%s', inserted nucleotides shouldn't map to reference",
                        read_name)
                    return None

        nucleotides_at_variant_locus = convert_from_bytes_if_necessary(
            sequence[read_base0_start_inclusive:read_base0_end_exclusive])

        if "N" in nucleotides_at_variant_locus:
            logger.debug(
                "Skipping read '%s', found N nucleotides at variant locus",
                read_name)
        prefix = convert_from_bytes_if_necessary(sequence[:read_base0_start_inclusive])
        suffix = convert_from_bytes_if_necessary(sequence[read_base0_end_exclusive:])

        prefix, suffix = trim_N_nucleotides(prefix, suffix)

        return AlleleRead(
            prefix,
            nucleotides_at_variant_locus,
            suffix,
            name=read_name)



from isovar.dataframe_helpers import allele_counts_dataframe
from isovar.allele_read import AlleleRead
from isovar.read_evidence import ReadEvidence
from varcode import Variant
from nose.tools import eq_


def test_allele_count_dataframe():
    variant = Variant("test_contig", 50, "C", "G")
    read_evidence = ReadEvidence(
            trimmed_base1_start=50,
            trimmed_ref="C",
            trimmed_alt="G",
            ref_reads=[
                AlleleRead(prefix="AAA", allele="C", suffix="TTT", name="C1"),
                AlleleRead(prefix="AAC", allele="C", suffix="TTA", name="C2"),
            ],
            alt_reads=[
                AlleleRead(prefix="AAA", allele="G", suffix="TTT", name="G1")
            ],
            other_reads=[])
    df = allele_counts_dataframe([(variant, read_evidence)])
    assert len(df) == 1, "Wrong number of rows in DataFrame: %s" % (df,)
    row = df.iloc[0]
    eq_(row.num_ref_reads, 2)
    eq_(row.num_alt_reads, 1)
    eq_(row.num_other_reads, 0)


if __name__ == "__main__":
    test_allele_count_dataframe()



from __future__ import print_function, division, absolute_import


from isovar import ProteinSequenceCreator, ReadCollector

from nose.tools import eq_

from testing_helpers import load_bam, load_vcf


def test_translate_variant_collection():
    variants = load_vcf("data/b16.f10/b16.vcf")
    samfile = load_bam("data/b16.f10/b16.combined.sorted.bam")
    read_evidence_gen = ReadCollector().read_evidence_generator(
        variants,
        samfile)
    translation_gen = ProteinSequenceCreator().translate_variants(read_evidence_gen)
    translations = list(translation_gen)
    eq_(
        len(translations),
        4,
        "Expected %d translated variants but got %d: %s" % (
            len(variants),
            len(translations),
            translations))



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import numpy as np


from .value_object import ValueObject
from .logging import get_logger

logger = get_logger(__name__)


class VariantSequence(ValueObject):
    """
    Representation of a cDNA sequence containing a mutation
    """

    __slots__ = [
        # nucleotides before a variant
        "prefix",
        # nucleotide sequence of a variant
        "alt",
        # nucleotides after a variant
        "suffix",
        # since we often want to look at prefix+alt+suffix, let's cache it
        "sequence",
        # reads which were used to determine this sequences
        "reads"
    ]

    def __init__(self, prefix, alt, suffix, reads):
        self.prefix = prefix
        self.alt = alt
        self.suffix = suffix
        self.sequence = prefix + alt + suffix
        self.reads = frozenset(reads)

    def __len__(self):
        return len(self.sequence)

    def __str__(self):
        return (
            "VariantSequence("
            "prefix='%s', alt='%s', "
            "suffix='%s', sequence='%s', "
            "reads=<%d AlleleRead object%s>)") % (
                self.prefix,
                self.alt,
                self.suffix,
                self.sequence,
                len(self.reads),
                "s" if len(self.reads) == 0 or len(self.reads) > 1 else "")

    def __repr__(self):
        return str(self)

    @property
    def read_names(self):
        """
        Unique read names from which this VariantSequence was constructed.

        Returns set of str
        """
        return {r.name for r in self.reads}

    def contains(self, other):
        """
        Is the other VariantSequence a subsequence of this one?

        The two sequences must agree on the alt nucleotides, the prefix of the
        longer must contain the prefix of the shorter, and the suffix of the
        longer must contain the suffix of the shorter.
        """
        return (self.alt == other.alt and
                self.prefix.endswith(other.prefix) and
                self.suffix.startswith(other.suffix))

    def left_overlaps(self, other, min_overlap_size=1):
        """
        Does this VariantSequence overlap another on the left side?
        """

        if self.alt != other.alt:
            # allele must match!
            return False

        if len(other.prefix) > len(self.prefix):
            # only consider strings that overlap like:
            #   self: ppppAssss
            #   other:   ppAsssssss
            # which excludes cases where the other sequence has a longer
            # prefix
            return False
        elif len(other.suffix) < len(self.suffix):
            # similarly, we throw away cases where the other sequence is shorter
            # after the alt nucleotides than this sequence
            return False

        # is the other sequence a prefix of this sequence?
        # Example:
        # p1 a1 s1 = XXXXXXXX Y ZZZZZZ
        # p2 a2 s2 =       XX Y ZZZZZZZZZ
        # ...
        # then we can combine them into a longer sequence
        sequence_overlaps = (
            self.prefix.endswith(other.prefix) and
            other.suffix.startswith(self.suffix)
        )
        prefix_overlap_size = min(len(self.prefix), len(other.prefix))
        suffix_overlap_size = min(len(other.suffix), len(self.suffix))
        overlap_size = (
            prefix_overlap_size + suffix_overlap_size + len(self.alt))

        return sequence_overlaps and overlap_size >= min_overlap_size

    def add_reads(self, reads):
        """
        Create another VariantSequence with more supporting reads.
        """
        if len(reads) == 0:
            return self
        new_reads = self.reads.union(reads)
        if len(new_reads) > len(self.reads):
            return VariantSequence(
                prefix=self.prefix,
                alt=self.alt,
                suffix=self.suffix,
                reads=new_reads)
        else:
            return self

    def combine(self, other_sequence, min_overlap_size=1):
        """
        If this sequence is the prefix of another sequence, combine
        them into a single VariantSequence object. If the other sequence
        is contained in this one, then add its reads to this VariantSequence.
        Also tries to flip the order (e.g. this sequence is a suffix or
        this sequence is a subsequence). If sequences can't be combined
        then returns None.
        """
        if other_sequence.alt != self.alt:
            logger.warn(
                "Cannot combine %s and %s with mismatching alt sequences",
                self,
                other_sequence)
            return None
        elif self.contains(other_sequence):
            if len(other_sequence) >= min_overlap_size:
                return self.add_reads(other_sequence.reads)
            else:
                return None
        elif other_sequence.contains(self):
            if len(self) >= min_overlap_size:
                return other_sequence.add_reads(self.reads)
            else:
                return None
        elif self.left_overlaps(other_sequence, min_overlap_size=min_overlap_size):
            # If sequences are like AABC and ABCC
            return VariantSequence(
                prefix=self.prefix,
                alt=self.alt,
                suffix=other_sequence.suffix,
                reads=self.reads.union(other_sequence.reads))
        elif other_sequence.left_overlaps(self, min_overlap_size=min_overlap_size):
            return VariantSequence(
                prefix=other_sequence.prefix,
                alt=self.alt,
                suffix=self.suffix,
                reads=self.reads.union(other_sequence.reads))
        else:
            # sequences don't overlap
            return None

    def variant_indices(self):
        """
        When we combine prefix + alt + suffix into a single string,
        what are is base-0 index interval which gets us back the alt
        sequence? First returned index is inclusive, the second is exclusive.
        """
        variant_start_index = len(self.prefix)
        variant_len = len(self.alt)
        variant_end_index = variant_start_index + variant_len
        return variant_start_index, variant_end_index

    def coverage(self):
        """
        Returns NumPy array indicating number of reads covering each
        nucleotides of this sequence.
        """
        variant_start_index, variant_end_index = self.variant_indices()
        n_nucleotides = len(self)
        coverage_array = np.zeros(n_nucleotides, dtype="int32")
        for read in self.reads:
            coverage_array[
                max(0, variant_start_index - len(read.prefix)):
                min(n_nucleotides, variant_end_index + len(read.suffix))] += 1
        return coverage_array

    def min_coverage(self):
        """
        Minimum number of reads covering any base in the cDNA sequence

        Returns int
        """
        return np.min(self.coverage())

    def mean_coverage(self):
        """
        Average number of reads covering each base in the cDNA sequence.

        Returns float
        """
        return np.mean(self.coverage())

    def trim_by_coverage(self, min_reads):
        """
        Given the min number of reads overlapping each nucleotide of
        a variant sequence, trim this sequence by getting rid of positions
        which are overlapped by fewer reads than specified.
        """
        read_count_array = self.coverage()
        logger.info("Coverage: %s (len=%d)" % (
            read_count_array, len(read_count_array)))
        sufficient_coverage_mask = read_count_array >= min_reads
        sufficient_coverage_indices = np.argwhere(sufficient_coverage_mask)
        if len(sufficient_coverage_indices) == 0:
            logger.debug("No bases in %s have coverage >= %d" % (self, min_reads))
            return VariantSequence(prefix="", alt="", suffix="", reads=self.reads)
        variant_start_index, variant_end_index = self.variant_indices()
        # assuming that coverage drops off monotonically away from
        # variant nucleotides
        first_covered_index = sufficient_coverage_indices.min()
        last_covered_index = sufficient_coverage_indices.max()
        # adding 1 to last_covered_index since it's an inclusive index
        # whereas variant_end_index is the end of a half-open interval
        if (first_covered_index > variant_start_index or
                last_covered_index + 1 < variant_end_index):
            # Example:
            #   Nucleotide sequence:
            #       ACCCTTTT|AA|GGCGCGCC
            #   Coverage:
            #       12222333|44|33333211
            # Then the mask for bases covered >= 4x would be:
            #       ________|**|________
            # with indices:
            #       first_covered_index = 9
            #       last_covered_index = 10
            #       variant_start_index = 9
            #       variant_end_index = 11
            logger.debug("Some variant bases in %s don't have coverage >= %d" % (
                self, min_reads))
            return VariantSequence(prefix="", alt="", suffix="", reads=self.reads)
        return VariantSequence(
            prefix=self.prefix[first_covered_index:],
            alt=self.alt,
            suffix=self.suffix[:last_covered_index - variant_end_index + 1],
            reads=self.reads)




# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Sequence alignment helpers
"""

from __future__ import print_function, division, absolute_import


def alignment_score(a, b, min_subsequence_length=1):
    """
    Number of mismatches between all two input sequences, allows
    for trimming of ends of sequences but not insertions or deletions
    within the sequences. Number of trimmed amino acids from each sequence
    count toward the mismatch total.

    Parameters
    ----------
    a : str

    b : str

    min_subsequence_length : int
        Only consider subsequences which are at least this long.
    Returns int
    """

    n_a = len(a)
    n_b = len(b)

    # swap a and b if a is longer since the loops below expect the first
    # sequence to be shorter. This happens because any subsequence of `a`
    # is expected to be of a length that can be found within `b`
    if n_a > n_b:
        a, b = b, a
        n_a, n_b = n_b, n_a

    # compare all subsequences of a and b and count the number of mismatches
    # between
    best_score = n_a + n_b

    # if we need to make sequence of at least length `min_subsequence_length`
    # then we can't start the sequence more than that number of characters
    # from the end of the string.
    # For example, if n_a = 7 and min_subsequence_length = 6 then
    # this loop should only consider start_a = {0, 1} but not any value
    # higher than that.
    for start_a in range(n_a - min_subsequence_length + 1):
        # Similarly, only consider end indices for the subsequence of `a`
        # which span the minimum number of characters required.
        # For example, if n_a = 7, min_subsequence_length = 6, start_a = 1
        # then this loop should only consider end_a = 7.
        for end_a in range(start_a + min_subsequence_length, n_a + 1):
            subseq_a = a[start_a:end_a]
            n_subseq_a = len(subseq_a)
            n_trimmed_a = n_a - n_subseq_a
            # consider all subsequences of the second string of the same length
            # as the subsequence extracted from the first string
            for start_b in range(n_b - n_subseq_a + 1):
                subseq_b = b[start_b:start_b + n_subseq_a]
                n_subseq_b = len(subseq_b)
                assert n_subseq_a == n_subseq_b
                n_trimmed_b = n_b - n_subseq_b
                # now that we have two subsequences of the same length,
                # count up the number of mismatching characters between them
                n_mismatches = sum([
                    a_char != b_char
                    for (a_char, b_char)
                    in zip(subseq_a, subseq_b)])
                score = n_trimmed_a + n_trimmed_b + n_mismatches
                if score < best_score:
                    best_score = score
    return best_score



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .reference_coding_sequence_key import ReferenceCodingSequenceKey
from .logging import get_logger

logger = get_logger(__name__)


##########################
#
# ReferenceContext
# ----------------
#
# Includes all the fields of ReferenceCodingSequenceKey in addition to which
# variant we're examining and all transcripts overlapping that variant
# which produced this particular sequence context and reading frame.
#
##########################

class ReferenceContext(ReferenceCodingSequenceKey):
    """
    Representation of the sequence preceding a variant and its associated
    reading frame.
    """

    # additional fields on top of slots for ReferenceCodingSequenceKey
    __slots__ = ["variant", "transcripts"]

    def __init__(
            self,
            strand,
            sequence_before_variant_locus,
            sequence_at_variant_locus,
            sequence_after_variant_locus,
            offset_to_first_complete_codon,
            contains_start_codon,
            overlaps_start_codon,
            contains_five_prime_utr,
            amino_acids_before_variant,
            variant,
            transcripts):
        ReferenceCodingSequenceKey.__init__(
            self,
            strand=strand,
            sequence_before_variant_locus=sequence_before_variant_locus,
            sequence_at_variant_locus=sequence_at_variant_locus,
            sequence_after_variant_locus=sequence_after_variant_locus,
            offset_to_first_complete_codon=offset_to_first_complete_codon,
            contains_start_codon=contains_start_codon,
            overlaps_start_codon=overlaps_start_codon,
            contains_five_prime_utr=contains_five_prime_utr,
            amino_acids_before_variant=amino_acids_before_variant)
        self.variant = variant
        self.transcripts = tuple(transcripts)

    @classmethod
    def from_reference_coding_sequence_key(cls, key, variant, transcripts):
        """
        Construct a ReferenceContext object from a ReferenceSequenceKey, variant,
        and a set of transcript.

        Parameters
        ----------
        key : ReferenceSequenceKey
        variant : varcode.Variant
        transcripts : list of pyensembl.Transcript

        Returns ReferenceContext
        """
        return ReferenceContext(
            strand=key.strand,
            sequence_before_variant_locus=key.sequence_before_variant_locus,
            sequence_at_variant_locus=key.sequence_at_variant_locus,
            sequence_after_variant_locus=key.sequence_after_variant_locus,
            offset_to_first_complete_codon=key.offset_to_first_complete_codon,
            contains_start_codon=key.contains_start_codon,
            overlaps_start_codon=key.overlaps_start_codon,
            contains_five_prime_utr=key.contains_five_prime_utr,
            amino_acids_before_variant=key.amino_acids_before_variant,
            variant=variant,
            transcripts=transcripts)

    def sort_key_decreasing_max_length_transcript_cds(self):
        """
        Used to sort a sequence of ReferenceContext objects by the longest CDS
        in each context's list of transcripts.
        """
        return -max(len(t.coding_sequence) for t in self.transcripts)

    @property
    def mitochondrial(self):
        """
        Is this a reference context for a variant in the mitochondrial
        genome?
        """
        return self.variant.contig.lower() in {"chrm", "m", "chrmt", "mt"}




from isovar import run_isovar
from isovar import ProteinSequence
from varcode import Variant
from testing_helpers import data_path

from nose.tools import eq_


def test_isovar_result_property_types():
    for result in run_isovar(
            variants=data_path("data/b16.f10/b16.vcf"),
            alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam")):
        # variant
        assert type(result.variant) is Variant

        # counts of genes and transcripts from variant
        assert type(result.num_overlapping_genes) is int
        assert type(result.num_overlapping_coding_genes) is int
        assert type(result.num_overlapping_transcripts) is int
        assert type(result.num_overlapping_coding_transcripts) is int

        # protein sequence
        assert type(result.top_protein_sequence) in (type(None), ProteinSequence)

        # counts of genes and transcripts from protein sequences
        assert type(result.num_genes_from_protein_sequences) is int
        assert type(result.num_genes_from_top_protein_sequence) is int
        assert type(result.num_transcripts_from_protein_sequences) is int
        assert type(result.num_transcripts_from_top_protein_sequence) is int

        # read and fragment counts
        assert type(result.num_ref_reads) is int
        assert type(result.num_alt_reads) is int
        assert type(result.num_other_reads) is int
        assert type(result.num_ref_fragments) is int
        assert type(result.num_alt_fragments) is int
        assert type(result.num_other_fragments) is int

        # read and fragment fractions
        assert type(result.fraction_ref_reads) is float
        assert type(result.fraction_alt_reads) is float
        assert type(result.fraction_other_reads) is float
        assert type(result.fraction_ref_fragments) is float
        assert type(result.fraction_alt_fragments) is float
        assert type(result.fraction_other_fragments) is float

        # read and fragment count ratios
        assert type(result.ratio_alt_to_other_reads) is float
        assert type(result.ratio_alt_to_other_fragments) is float
        assert type(result.ratio_other_to_alt_fragments) is float
        assert type(result.ratio_other_to_alt_reads) is float
        assert type(result.ratio_ref_to_other_fragments) is float
        assert type(result.ratio_other_to_ref_fragments) is float
        assert type(result.ratio_other_to_ref_reads) is float

        # this property aggregates all filters
        assert result.passes_all_filters in {True, False}

        assert type(result.protein_sequence_mutation_start) in (int, type(None))
        assert type(result.protein_sequence_mutation_end) in (int, type(None))


def test_isovar_result_nonsyn_variants():
    for result in run_isovar(
            variants=data_path("data/b16.f10/b16.vcf"),
            alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam")):
        print(result.variant)
        print(result.predicted_effect)
        if result.has_mutant_protein_sequence_from_rna:
            assert result.num_amino_acid_mismatches_from_predicted_effect is not None
            assert result.num_amino_acid_mismatches_from_reference is not None
            assert result.num_amino_acid_mismatches_from_reference > 0
            eq_(result.protein_sequence_matches_reference, False)
            eq_(result.protein_sequence_contains_mutation, True)
        else:
            assert result.num_amino_acid_mismatches_from_predicted_effect is None
            assert result.num_amino_acid_mismatches_from_reference is None
            assert result.protein_sequence_matches_predicted_mutation_effect is None
            assert result.protein_sequence_matches_reference is None
            assert result.protein_sequence_contains_mutation is None


def test_isovar_result_clone():
    for result in run_isovar(
            variants=data_path("data/b16.f10/b16.vcf"),
            alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam")):
        result2 = result.clone()
        eq_(result, result2)

def test_isovar_result_clone_with_updates():
    for result in run_isovar(
            variants=data_path("data/b16.f10/b16.vcf"),
            alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam")):
        result2 = result.clone_with_updates(variant=None)
        assert result != result2


def test_isovar_result_str():
    for result in run_isovar(
            variants=data_path("data/b16.f10/b16.vcf"),
            alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam")):
        s = str(result)
        assert len(s) > 0
        assert s.startswith("IsovarResult(")
        assert s.endswith(")")


from __future__ import print_function, division, absolute_import

from isovar import ReadCollector
from isovar.allele_read_helpers import  group_unique_sequences

from varcode import Variant

from testing_helpers import load_bam
from genomes_for_testing import grch38


def test_group_unique_sequences():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 65857041
    ref = "G"
    alt = "C"
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref, alt=alt,
        ensembl=grch38)
    read_collector = ReadCollector()
    variant_reads = read_collector.allele_reads_supporting_variant(
        alignment_file=samfile,
        variant=variant)
    print("%d variant reads: %s" % (
        len(variant_reads), variant_reads))
    groups = group_unique_sequences(
        variant_reads,
        max_prefix_size=30,
        max_suffix_size=30)
    print("%d unique sequences: %s" % (
        len(groups), groups))
    # there are some redundant reads, so we expect that the number of
    # unique entries should be less than the total read partitions
    assert len(variant_reads) > len(groups)



from __future__ import print_function, division, absolute_import

from nose.tools import eq_
from varcode import Variant

from isovar.allele_read import AlleleRead
from isovar.read_collector import ReadCollector

from mock_objects import MockAlignmentFile, make_pysam_read
from genomes_for_testing import grch38


def test_partitioned_read_sequences_snv():
    """
    test_partitioned_read_sequences_snv : Test that read gets correctly
    partitioned for chr1:4 T>G where the sequence for chr1 is assumed
    to be "ACCTTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "T"
    alt = "G"

    variant = Variant(
        chromosome,
        location,
        ref,
        alt,
        grch38,
        normalize_contig_name=False)

    read = make_pysam_read(seq="ACCGTG", cigar="6M", mdtag="3G2")

    samfile = MockAlignmentFile(
        references=(chromosome,),
        reads=[read])
    read_creator = ReadCollector()
    variant_reads = read_creator.allele_reads_supporting_variant(
        variant=variant,
        alignment_file=samfile)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = AlleleRead(
        name=read.qname,
        prefix="ACC",
        allele="G",
        suffix="TG")
    eq_(variant_read, expected)


def test_partitioned_read_sequences_insertion():
    """
    test_partitioned_read_sequences_insertion : Test that read gets correctly
    partitioned for chr1:4 T>TG
    where the sequence for chr1 is assumed to be "ACCTTG"
    and the variant sequence is "ACCTGTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "T"
    alt = "TG"
    variant = Variant(
        chromosome, location, ref, alt, grch38, normalize_contig_name=False)

    read = make_pysam_read(
        seq=b"ACCTGTG",
        cigar="4M1I2M",
        mdtag="6")

    samfile = MockAlignmentFile(
        references=(chromosome,),
        reads=[read])
    read_creator = ReadCollector()

    variant_reads = read_creator.allele_reads_supporting_variant(
        alignment_file=samfile,
        variant=variant)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = AlleleRead(
        name=read.qname,
        prefix="ACCT",
        allele="G",
        suffix="TG")
    eq_(variant_read, expected)


def test_partitioned_read_sequences_deletion():
    """
    test_partitioned_read_sequences_deletion : Test that read gets correctly
    partitioned for chr1:4 TT>T where the sequence for chr1 is assumed to
    be "ACCTTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "TT"
    alt = "T"
    variant = Variant(
        chromosome, location, ref, alt, grch38, normalize_contig_name=False)

    read = make_pysam_read(
        seq="ACCTG",
        cigar="4M1D1M",
        mdtag="4^T1")
    samfile = MockAlignmentFile(
        references=(chromosome,),
        reads=[read])
    read_creator = ReadCollector()
    variant_reads = read_creator.allele_reads_supporting_variant(
        alignment_file=samfile,
        variant=variant)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = AlleleRead(
        name=read.qname,
        prefix="ACCT",
        allele="",
        suffix="G")
    eq_(variant_read, expected)



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Helper functions used for creating translating a variant's cDNA sequence
into a particular reading frame.
"""

from __future__ import print_function, division, absolute_import

import math

def find_mutant_amino_acid_interval(
        cdna_sequence,
        cdna_first_codon_offset,
        cdna_variant_start_offset,
        cdna_variant_end_offset,
        n_ref,
        n_amino_acids):
    """
    Parameters
    ----------
    cdna_sequence : skbio.DNA or str
        cDNA sequence found in RNAseq data

    cdna_first_codon_offset : int
        Offset into cDNA sequence to first complete codon, lets us skip
        past UTR region and incomplete codons.

    cdna_variant_start_offset : int
        Interbase start offset into cDNA sequence for selecting mutant
        nucleotides.

    cdna_variant_end_offset : int
        Interbase end offset into cDNA sequence for selecting mutant
        nucleotides.

    n_ref : int
        Number of reference nucleotides

    n_amino_acids : int
        Number of translated amino acids

    Returns
    -------
    tuple with three fields:
        1) Start offset for interval of mutant amino acids in translated sequence
        2) End offset for interval of mutant amino acids in translated sequence
        3) Boolean flag indicating whether the variant was a frameshift.
    """
    cdna_alt_nucleotides = cdna_sequence[
        cdna_variant_start_offset:cdna_variant_end_offset]

    n_alt = len(cdna_alt_nucleotides)

    # sequence of nucleotides before the variant starting from the first codon
    cdna_coding_prefix = cdna_sequence[cdna_first_codon_offset:cdna_variant_start_offset]

    # rounding down since a change in the middle of a codon should count
    # toward the variant codons
    n_coding_nucleotides_before_variant = len(cdna_coding_prefix)

    n_complete_prefix_codons = n_coding_nucleotides_before_variant // 3

    frame_of_variant_nucleotides = n_coding_nucleotides_before_variant % 3
    frameshift = abs(n_ref - n_alt) % 3 != 0
    indel = n_ref != n_alt

    variant_aa_interval_start = n_complete_prefix_codons

    if frameshift:
        # if mutation is a frame shift then every amino acid from the
        # first affected codon to the stop is considered mutant
        #
        # TODO: what if the first k amino acids are synonymous with the reference sequence?
        variant_aa_interval_end = n_amino_acids
    else:
        n_alt_codons = int(math.ceil(n_alt / 3.0))
        if indel:
            # We need to adjust the number of affected codons by whether the
            # variant is aligned with codon boundaries, since in-frame indels
            # may still be split across multiple codons.
            #
            # Example of in-frame deletion of 3 nucleotides which leaves
            # 0 variant codons in the sequence (interval = 1:1)
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CCC|GGG|TTT
            #
            # Example of in-frame deletion of 3 nucleotides which leaves
            # 1 variant codon in the sequence (interval = 1:2)
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CCC|AGG|TTT
            #
            # Example of in-frame insertion of 3 nucleotides which
            # yields two variant codons:
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CTT|TCC|AAA|GGG|TTT
            extra_affected_codon = int(frame_of_variant_nucleotides != 0)
            variant_aa_interval_end = (
                variant_aa_interval_start + n_alt_codons + extra_affected_codon)
        else:
            # if the variant is a simple substitution then it only affects
            # as many codons as are in the alternate sequence
            variant_aa_interval_end = variant_aa_interval_start + n_alt_codons
    return variant_aa_interval_start, variant_aa_interval_end, frameshift



# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from .default_parameters import (
    MIN_TRANSCRIPT_PREFIX_LENGTH,
    MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
    COUNT_MISMATCHES_AFTER_VARIANT,
    PROTEIN_SEQUENCE_LENGTH,
    MAX_PROTEIN_SEQUENCES_PER_VARIANT,
    MIN_VARIANT_SEQUENCE_COVERAGE,
    VARIANT_SEQUENCE_ASSEMBLY,
    MIN_VARIANT_SEQUENCE_ASSEMBLY_OVERLAP_SIZE,
)

from .genetic_code import translate_cdna
from .protein_sequence_helpers import (
    sort_protein_sequences,
    group_equivalent_translations
)
from .reference_context_helpers import reference_contexts_for_variant
from .translation import Translation
from .translation_helpers import find_mutant_amino_acid_interval
from .value_object import ValueObject
from .variant_sequence_creator import VariantSequenceCreator
from .variant_orf_helpers import match_variant_sequence_to_reference_context

from .logging import get_logger

logger = get_logger(__name__)


class ProteinSequenceCreator(ValueObject):
    """
    Creates ProteinSequence objects for each variant by translating
    cDNA into one or more Translation objects and then grouping them
    by identical amino acid sequences. Each Translation comes from the
    combination of a variant cDNA sequence with the reading frames of
    annotated reference transcripts.
    """

    def __init__(
            self,
            protein_sequence_length=PROTEIN_SEQUENCE_LENGTH,
            min_variant_sequence_coverage=MIN_VARIANT_SEQUENCE_COVERAGE,
            min_transcript_prefix_length=MIN_TRANSCRIPT_PREFIX_LENGTH,
            max_transcript_mismatches=MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
            count_mismatches_after_variant=COUNT_MISMATCHES_AFTER_VARIANT,
            max_protein_sequences_per_variant=MAX_PROTEIN_SEQUENCES_PER_VARIANT,
            variant_sequence_assembly=VARIANT_SEQUENCE_ASSEMBLY,
            min_assembly_overlap_size=MIN_VARIANT_SEQUENCE_ASSEMBLY_OVERLAP_SIZE):
        """
        protein_sequence_length : int
            Try to translate protein sequences of this length, though sometimes
            we'll have to return something shorter (depending on the RNAseq data,
            and presence of stop codons).

        min_variant_sequence_coverage : int
            Trim variant sequences to positions supported by at least this number
            of RNA reads.

        min_transcript_prefix_length : int
            Minimum number of bases we need to try matching between the reference
            context and variant sequence.

        max_transcript_mismatches : int
            Don't try to determine the reading frame for a transcript if more
            than this number of bases differ.

        count_mismatches_after_variant : bool
            Include mismatches after the variant locus in the count compared
            against max_transcript_mismatches.

        max_protein_sequences_per_variant : int
            Number of protein sequences to return for each ProteinSequence

        variant_sequence_assembly : bool
            If True, then assemble variant cDNA sequences based on overlap of
            RNA reads. If False, then variant cDNA sequences must be fully spanned
            and contained within RNA reads.

        min_assembly_overlap_size : int
            Minimum number of nucleotides that two reads need to overlap before they
            can be merged into a single coding sequence.
        """
        self.protein_sequence_length = protein_sequence_length
        self.min_variant_sequence_coverage = min_variant_sequence_coverage
        self.min_transcript_prefix_length = min_transcript_prefix_length
        self.max_transcript_mismatches = max_transcript_mismatches
        self.count_mismatches_after_variant = count_mismatches_after_variant
        self.variant_sequence_assembly = variant_sequence_assembly
        self.min_assembly_overlap_size = min_assembly_overlap_size

        # Adding an extra codon to the desired RNA sequence length in case we
        # need to clip nucleotides at the start/end of the sequence
        self._cdna_sequence_length = (self.protein_sequence_length + 1) * 3

        self._variant_sequence_creator = VariantSequenceCreator(
            min_variant_sequence_coverage=self.min_variant_sequence_coverage,
            preferred_sequence_length=self._cdna_sequence_length,
            variant_sequence_assembly=self.variant_sequence_assembly,
            min_assembly_overlap_size=self.min_assembly_overlap_size)

        self.max_protein_sequences_per_variant = max_protein_sequences_per_variant

    def translation_from_variant_sequence_and_reference_context(
                self,
                variant_sequence,
                reference_context):
        """
        Attempt to translate a single VariantSequence using the reading frame
        from a single ReferenceContext.

        Parameters
        ----------
        variant_sequence : VariantSequence

        reference_context : ReferenceContext

        Returns either a Translation object or None if the number of
        mismatches between the RNA and reference transcript sequences exceeds
        given threshold.
        """

        logger.info(
            "Full mutant cDNA sequence: %s (len=%d)",
            variant_sequence.sequence,
            len(variant_sequence))
        variant_orf = match_variant_sequence_to_reference_context(
            variant_sequence,
            reference_context,
            min_transcript_prefix_length=self.min_transcript_prefix_length,
            max_transcript_mismatches=self.max_transcript_mismatches,
            count_mismatches_after_variant=self.count_mismatches_after_variant)

        if variant_orf is None:
            logger.info("Unable to determine reading frame for %s", variant_sequence)
            return None

        cdna_sequence = variant_orf.cdna_sequence
        cdna_codon_offset = variant_orf.offset_to_first_complete_codon
        logger.info(
            "Untrimmed cDNA sequence: %s, offset to first codon = %d, len=%d",
            cdna_sequence,
            cdna_codon_offset,
            len(cdna_sequence))
        # get the offsets into the cDNA sequence which pick out the variant nucleotides
        cdna_variant_start_offset = variant_orf.variant_cdna_interval_start
        cdna_variant_end_offset = variant_orf.variant_cdna_interval_end

        in_frame_cdna_sequence = cdna_sequence[cdna_codon_offset:]
        logger.info("Translating '%s' (len=%d, expected AA length=%d)",
            in_frame_cdna_sequence,
            len(in_frame_cdna_sequence),
            len(in_frame_cdna_sequence) // 3)
        # TODO:
        #  determine if the first codon is the start codon of a
        #  transcript, for now any of the unusual start codons like CTG
        #  will translate to leucine instead of methionine
        variant_amino_acids, ends_with_stop_codon = translate_cdna(
            in_frame_cdna_sequence,
            first_codon_is_start=False,
            mitochondrial=reference_context.mitochondrial)
        logger.info("Mutant amino acids: %s, ends_with_stop=%s, len=%d" % (
            variant_amino_acids,
            ends_with_stop_codon,
            len(variant_amino_acids)))
        variant_aa_interval_start, variant_aa_interval_end, frameshift = \
            find_mutant_amino_acid_interval(
                cdna_sequence=cdna_sequence,
                cdna_first_codon_offset=cdna_codon_offset,
                cdna_variant_start_offset=cdna_variant_start_offset,
                cdna_variant_end_offset=cdna_variant_end_offset,
                n_ref=len(reference_context.sequence_at_variant_locus),
                n_amino_acids=len(variant_amino_acids))

        if self.protein_sequence_length:
            if len(variant_amino_acids) > self.protein_sequence_length:
                if self.protein_sequence_length <= variant_aa_interval_start:
                    logger.warn(
                        ("Truncating amino acid sequence %s "
                         "to only %d elements loses all variant residues"),
                        variant_amino_acids,
                        self.protein_sequence_length)
                    return None
                else:
                    # if the protein is too long then shorten it, which implies
                    # we're no longer stopping due to a stop codon and that the variant
                    # amino acids might need a new stop index
                    variant_amino_acids = variant_amino_acids[:self.protein_sequence_length]
                    variant_aa_interval_end = min(
                        variant_aa_interval_end,
                        self.protein_sequence_length)
                    ends_with_stop_codon = False

        translation = Translation(
            amino_acids=variant_amino_acids,
            frameshift=frameshift,
            ends_with_stop_codon=ends_with_stop_codon,
            variant_aa_interval_start=variant_aa_interval_start,
            variant_aa_interval_end=variant_aa_interval_end,
            untrimmed_variant_sequence=variant_sequence,
            reference_context=reference_context,
            variant_orf=variant_orf)

        logger.info(
            ("Translation from:"
             "\n-- cDNA = %s"
             "\n-- context = %s"
             "\n-- translation = %s") % (
                variant_sequence,
                reference_context,
                translation))

        return translation

    def all_pairs_translations(
            self,
            variant_sequences,
            reference_contexts):
        """
        Given all a list of VariantSequence objects for a particular variant
        and all the ReferenceContext objects for that locus, attempt to
        translate all pairs of sequences and reference contexts.

        Parameters
        ----------
        variant_sequences : list of VariantSequence

        reference_contexts : list of ReferenceContext

        Return list of Translation objects.
        """
        translations = []
        for reference_context in reference_contexts:
            for variant_sequence in variant_sequences:
                translation = self.translation_from_variant_sequence_and_reference_context(
                    variant_sequence=variant_sequence,
                    reference_context=reference_context)
                if translation is not None:
                    translations.append(translation)
        return translations

    def translate_variant_reads(
            self,
            variant,
            variant_reads,
            transcript_id_whitelist=None):
        """
        Given a variant and its associated alt reads, construct variant sequences
        and translate them into Translation objects.

        Returns 0 or more Translation objects.

        Parameters
        ----------
        variant : varcode.Variant

        variant_reads : sequence or generator
            AlleleRead objects supporting the variant

        transcript_id_whitelist : set, optional
            If given, expected to be a set of transcript IDs which we should use
            for determining the reading frame around a variant. If omitted, then
            try to use all overlapping reference transcripts.

        Returns list of Translation objects
        """
        if len(variant_reads) == 0:
            logger.info("No supporting reads for variant %s", variant)
            return []

        reference_contexts = reference_contexts_for_variant(
            variant,
            context_size=self._cdna_sequence_length,
            transcript_id_whitelist=transcript_id_whitelist)

        if len(reference_contexts) == 0:
            logger.info("Could not determine reference context for variant %s", variant)
            return []

        variant_sequences = self._variant_sequence_creator.reads_to_variant_sequences(
            variant=variant,
            reads=variant_reads)

        if not variant_sequences:
            logger.info("No spanning cDNA sequences for variant %s", variant)
            return []

        return self.all_pairs_translations(
            variant_sequences=variant_sequences,
            reference_contexts=reference_contexts)

    def translate_variants(
                self,
                variants_with_read_evidence_generator,
                transcript_id_whitelist=None):
            """
            Translates each coding variant in a collection to one or more protein
            fragment sequences (if the variant is not filtered and its spanning RNA
            sequences can be given a reading frame).

            Parameters
            ----------
            variants_with_read_evidence_generator : sequence or generator
                Each item of this sequence should be a pair containing a varcode.Variant
                and a ReadEvidence object

            transcript_id_whitelist : set, optional
                If given, expected to be a set of transcript IDs which we should use
                for determining the reading frame around a variant. If omitted, then
                try to use all overlapping reference transcripts.

            Yields pairs of a Variant and a sequence of all its candidate
            Translation objects.
            """
            for variant, read_evidence in variants_with_read_evidence_generator:
                translations = self.translate_variant_reads(
                    variant=variant,
                    variant_reads=read_evidence.alt_reads,
                    transcript_id_whitelist=transcript_id_whitelist)
                yield variant, translations

    def sorted_protein_sequences_for_variant(
            self,
            variant,
            read_evidence,
            transcript_id_whitelist=None):
        """"
        Translates a coding variant and its overlapping RNA reads into Translation
        objects, which are aggregated into ProteinSequence objects by their
        amino acid sequence (when they have equivalent coding sequences).

        Parameters
        ----------
        variant : varcode.Variant

        read_evidence : ReadEvidence object

        transcript_id_whitelist : set, optional
            If given, expected to be a set of transcript IDs which we should use
            for determining the reading frame around a variant. If omitted, then
            try to use all overlapping reference transcripts.

        Returns a list of ProteinSequence objects
        """
        translations = self.translate_variant_reads(
            variant=variant,
            variant_reads=read_evidence.alt_reads,
            transcript_id_whitelist=transcript_id_whitelist)

        # group distinct cDNA translations into ProteinSequence objects
        # by their amino acid sequence
        protein_sequences = group_equivalent_translations(translations)

        # sort protein sequences before returning the top results
        protein_sequences = sort_protein_sequences(protein_sequences)
        return protein_sequences

    def protein_sequences_from_read_evidence_generator(
            self,
            read_evidence_generator,
            transcript_id_whitelist=None):
        """

        Parameters
        ----------
        read_evidence_generator : generator of (varcode.Variant, ReadEvidence)
            Generator which yields sequence of Variant objects paired with
            their corresponding ReadEvidence

        transcript_id_whitelist : set of str or None
            Which transcripts should be considered when predicting DNA-only
            coding effects of mutations and also when trying to establish a
            reading frame for identified cDNA sequences.

        Generates sequence of (varcode.Variant, ProteinSequence list) pairs.
        """
        for variant, read_evidence in read_evidence_generator:
            protein_sequences = \
                self.sorted_protein_sequences_for_variant(
                    variant=variant,
                    read_evidence=read_evidence,
                    transcript_id_whitelist=transcript_id_whitelist)
            if self.max_protein_sequences_per_variant:
                protein_sequences = protein_sequences[:self.max_protein_sequences_per_variant]
            yield variant, protein_sequences



# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Since multiple variant sequences can translate to the same amino acid sequence,
this module aggregates equivalent Translation objects into a single
ProteinSequence.
"""

from __future__ import print_function, division, absolute_import

from .common import groupby
from .logging import get_logger
from .protein_sequence import ProteinSequence
from .translation import Translation

logger = get_logger(__name__)


def sort_protein_sequences(protein_sequences):
    """
    Sort protein sequences in decreasing order of priority
    """
    return list(
        sorted(
            protein_sequences,
            key=ProteinSequence.ascending_sort_key,
            reverse=True))


def group_equivalent_translations(translations):
    """
    Convert a list of Translation objects into a (potentially smaller) list
    of ProteinSequence objects by grouping the equivalent amino acid sequences.

    Parameters
    ----------
    translations : list of Translation objects

    Returns list of ProteinSequence objects
    """
    protein_sequences = []
    translation_groups = groupby(
        translations,
        key_fn=Translation.as_translation_key)
    for equivalent_translations in translation_groups.values():
        protein_sequences.append(ProteinSequence(equivalent_translations))
    return protein_sequences



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for filtering, grouping, and summarizing collections of
AlleleRead objects.
"""

from collections import defaultdict

from .common import groupby
from .logging import get_logger
from .allele_read import AlleleRead

logger = get_logger(__name__)


def group_reads_by_allele(allele_reads):
    """
    Returns dictionary mapping each allele's nucleotide sequence to a list of
    supporting AlleleRead objects.
    """
    return groupby(allele_reads, lambda read: read.allele)


def get_single_allele_from_reads(allele_reads):
    """
    Given a sequence of AlleleRead objects, which are expected to all have
    the same allele, return that allele.
    """
    allele_reads = list(allele_reads)

    if len(allele_reads) == 0:
        raise ValueError("Expected non-empty list of AlleleRead objects")

    seq = allele_reads[0].allele
    if any(read.allele != seq for read in allele_reads):
        raise ValueError("Expected all AlleleRead objects to have same allele '%s', got %s" % (
            seq, allele_reads))
    return seq


def group_unique_sequences(
        allele_reads,
        max_prefix_size=None,
        max_suffix_size=None):
    """
    Given a list of AlleleRead objects, extracts all unique
    (prefix, allele, suffix) sequences and associate each with a list
    of reads that contained that sequence.
    """
    groups = defaultdict(set)
    for r in allele_reads:
        prefix = r.prefix
        allele = r.allele
        suffix = r.suffix
        if max_prefix_size and len(prefix) > max_prefix_size:
            prefix = prefix[-max_prefix_size:]
        if max_suffix_size and len(suffix) > max_suffix_size:
            suffix = suffix[:max_suffix_size]
        key = (prefix, allele, suffix)
        groups[key].add(r)
    return groups


def allele_reads_from_locus_reads(locus_reads):
    """
    Attempt to convert each LocusRead object to an AlleleRead and return
    the successfully converted objects.

    Parameters
    ----------
    locus_reads : list of LocusRead

    Returns list of AlleleRead
    -------

    """
    allele_reads = []
    for locus_read in locus_reads:
        allele_read = AlleleRead.from_locus_read(locus_read)
        if allele_read is None:
            continue
        else:
            allele_reads.append(allele_read)
    return allele_reads


def split_reads_into_ref_alt_other(ref, alt, overlapping_reads):
    """
    Returns three lists of AlleleRead objects
        - reads which support the reference allele
        - reads which support the variant's alt allele
        - reads which support other alleles
    """
    # convert to list in case it's a generator since
    # we want to traverse the sequence repeatedly
    overlapping_reads = list(overlapping_reads)

    reads_grouped_by_allele = group_reads_by_allele(overlapping_reads)
    ref_reads = reads_grouped_by_allele.get(ref, [])
    alt_reads = reads_grouped_by_allele.get(alt, [])
    other_reads = []
    for allele, allele_reads in reads_grouped_by_allele.items():
        if allele in {ref, alt}:
            continue
        other_reads.extend(allele_reads)
    return ref_reads, alt_reads, other_reads


# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions used to annotate IsovarResult objects with filters.
"""

from __future__ import print_function, division, absolute_import

from collections import OrderedDict
import operator


def evaluate_threshold_filters(isovar_result, filter_thresholds):
    """
    Helper method used by apply_filters

    Parameters
    ----------
    isovar_result : IsovarResult

    filter_thresholds : dict or OrderedDict
        Every argument is supposed to be something like "max_alt_reads"
        where the first three characters are "min" or "max" and the
        rest of the name is either a field of IsovarResult or
        a numeric field like "num_alt_reads". The name of each filter
        maps to a cutoff value. Filters starting with "max"
        require that the corresponding field on CoverageStats
        is <= cutoff, whereas filters starting with
        "min" require >= cutoff.

    Returns OrderedDict
    """
    filter_values_dict = OrderedDict()
    for name, threshold in filter_thresholds.items():
        parts = name.split("_")
        min_or_max = parts[0]
        field_name = "_".join(parts[1:])
        if min_or_max == "min":
            comparison_fn = operator.ge
        elif min_or_max == "max":
            comparison_fn = operator.le
        else:
            raise ValueError(
                "Invalid filter '%s', must start with 'min' or 'max'" % name)
        if hasattr(isovar_result, field_name):
            field_value = getattr(isovar_result, field_name)
        else:
            raise ValueError(
                "Invalid filter '%s' IsovarResult does not have property '%s'" % (
                    name,
                    field_name))
        filter_values_dict[name] = comparison_fn(field_value, threshold)
    return filter_values_dict


def evaluate_boolean_filters(isovar_result, filter_flags):
    """
    Helper function used by apply_filters.

    Parameters
    ----------
    isovar_result : IsovarResult

    filter_flags : list of str
        Every element should be a boolean property of IsovarResult
        or "not_" and the name of a property to be negated.

    Returns OrderedDict
    """
    filter_values = OrderedDict()
    for boolean_filter_name in filter_flags:
        if boolean_filter_name.startswith("not_"):
            boolean_field_name = boolean_filter_name[4:]
            negate = True
        else:
            boolean_field_name = boolean_filter_name
            negate = False
        if hasattr(isovar_result, boolean_field_name):
            field_value = getattr(isovar_result, boolean_field_name)
        else:
            raise ValueError(
                "IsovarResult does not have field name '%s'" % boolean_field_name)
        if field_value is None:
            field_value = False
        elif field_value not in {True, False}:
            raise ValueError("Expected filter '%s' to be boolean but got %s" % (
                boolean_filter_name,
                field_value))
        filter_values[boolean_filter_name] = (
            not field_value if negate else field_value
        )
    return filter_values


def evaluate_filters(
        isovar_result,
        filter_thresholds,
        filter_flags=[]):
    """
    Creates a dictionary whose keys are named of different
    filter conditions and values are booleans, where True
    indicates whether this set of coverage stats passes
    the filter and False indicates that it failed.

    Parameters
    ----------
    isovar_result : IsovarResult

    filter_thresholds : dict or OrderedDict
        Every argument is supposed to be something like "max_alt_reads"
        where the first three characters are "min" or "max" and the
        rest of the name is either a field of IsovarResult or
        a numeric field like "num_alt_reads". The name of each filter
        maps to a cutoff value. Filters starting with "max"
        require that the corresponding field on CoverageStats
        is <= cutoff, whereas filters starting with
        "min" require >= cutoff.

    filter_flags : list of str
        Every element should be a boolean property of IsovarResult
        or "not_" and the name of a property to be negated.

    Returns
    -------
    Dictionary of filter names mapped to boolean value indicating
    whether this locus passed the filter.
    """
    filter_values_dict = evaluate_boolean_filters(isovar_result, filter_flags)
    filter_values_dict.update(
        evaluate_threshold_filters(isovar_result, filter_thresholds))
    return filter_values_dict


def apply_filters(
        isovar_result,
        filter_thresholds={},
        filter_flags=[]):
    """
    Given an IsovarResult object, evaluates given filters
    for each object, and returns a copy of the IsovarResult with new fiter
    values.

    Parameters
    ----------
    isovar_result : IsovarResult

    filter_thresholds : dict or OrderedDict
        Every argument is supposed to be something like "max_alt_reads"
        where the first three characters are "min" or "max" and the
        rest of the name is either a field of IsovarResult or
        a numeric field like "num_alt_reads". The name of each filter
        maps to a cutoff value. Filters starting with "max"
        require that the corresponding field on CoverageStats
        is <= cutoff, whereas filters starting with
        "min" require >= cutoff.

    filter_flags : list of str
        Every element should be a boolean property of IsovarResult
        or "not_" and the name of a property to be negated.

    Returns IsovarResult
    """
    filter_values = OrderedDict(isovar_result.filter_values.items())
    new_filter_values = evaluate_filters(
        isovar_result,
        filter_thresholds=filter_thresholds,
        filter_flags=filter_flags)
    filter_values.update(new_filter_values)
    return isovar_result.clone_with_updates(filter_values=filter_values)



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module combines variant cDNA sequences collected from a BAM file with
the reading frames of annotated reference transcripts to create candidate
translations.
"""


from __future__ import print_function, division, absolute_import

from .logging import get_logger
from .value_object import ValueObject


logger = get_logger(__name__)


class TranslationKey(ValueObject):
    """
    TranslationKey contains fields related to a translated protein sequence
    which should be used to combine multiple equivalent mutated amino acid
    sequences.
    """
    __slots__ = [
        # translated sequence of a variant sequence in the ORF established
        # by a reference context
        "amino_acids",
        # half-open interval coordinates for variant amino acids
        # in the translated sequence
        "variant_aa_interval_start",
        "variant_aa_interval_end",
        # did the amino acid sequence end due to a stop codon or did we
        # just run out of sequence context around the variant?
        "ends_with_stop_codon",
        # was the variant a frameshift relative to the reference sequence?
        "frameshift"
    ]



# Copyright (c) 2018-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Collect AlleleReads overlapping each variant grouped by whether
they support the reference, somatic allele, or some other allele.
"""

from __future__ import print_function, division, absolute_import

from .allele_read_helpers import split_reads_into_ref_alt_other
from .variant_helpers import trim_variant
from .value_object import ValueObject


class ReadEvidence(ValueObject):
    """
    This class represents the reads at a variant locus partitioned
    by allele (ref/alt/other) relative to a variant.
    """

    __slots__ = [
        "trimmed_base1_start",
        "trimmed_ref",
        "trimmed_alt",
        "ref_reads",
        "alt_reads",
        "other_reads"
    ]

    @classmethod
    def from_variant_and_allele_reads(
            cls,
            variant,
            allele_reads):
        """
        Create a GroupedAlleleReads object from a variant and the set of reads overlapping
        the location of that variant.

        Parameters
        ----------
        variant : varcode.Variant

        allele_reads : list of AlleleRead

        Returns GroupedAlleleReads

        """
        trimmed_base1_start, trimmed_ref, trimmed_alt = \
            trim_variant(variant)
        ref_reads, alt_reads, other_reads = split_reads_into_ref_alt_other(
            ref=trimmed_ref,
            alt=trimmed_alt,
            overlapping_reads=allele_reads)
        return cls(
            trimmed_base1_start=trimmed_base1_start,
            trimmed_ref=trimmed_ref,
            trimmed_alt=trimmed_alt,
            ref_reads=ref_reads,
            alt_reads=alt_reads,
            other_reads=other_reads)

    def __init__(
            self,
            trimmed_base1_start,
            trimmed_ref,
            trimmed_alt,
            ref_reads,
            alt_reads,
            other_reads):
        """
        Parameters
        ----------
        trimmed_base1_start : int
            Base-1 inclusive genomic position indicating start of variant,
            trimmed to avoid any shared prefix between the ref and alt sequences

        trimmed_ref : str
            Reference allele, trimmed to avoid any shared prefix with the alt
            sequence.

        trimmed_alt : str
            Alternate allele, trimmed to avoid any shared prefix with the ref
            sequence.

        ref_reads : list of AlleleRead
            Reads supporting the reference allele

        alt_reads : list of AlleleRead
            Reads supporting the alt allele

        other_reads : list of AlleleRead
            Reads supporting some allele other than ref or alt.
        """
        self.trimmed_base1_start = trimmed_base1_start
        self.trimmed_ref = trimmed_ref
        self.trimmed_alt = trimmed_alt
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads
        self.other_reads = other_reads

    @property
    def ref_read_names(self):
        """
        Names of reads which match the ref allele

        Returns set of str
        """
        return {r.name for r in self.ref_reads}

    @property
    def alt_read_names(self):
        """
        Names of reads which match the alt allele

        Returns set of str
        """
        return {r.name for r in self.alt_reads}

    @property
    def other_read_names(self):
        """
        Names of reads which match non-ref/non-alt alleles

        Returns set of str
        """
        return {r.name for r in self.other_reads}



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module wraps pysam and gives us a view of any reads overlapping
a variant locus which includes offsets into the read sequence & qualities
for extracting variant nucleotides.
"""

from __future__ import print_function, division, absolute_import

from .value_object import ValueObject


class LocusRead(ValueObject):
    """
    Minimal set of information extracted from SAM/BAM alignment file at a particular
    locus to later figure out the allele at this locus.
    """
    __slots__ = [
        "name",
        "sequence",
        "reference_positions",
        "quality_scores",
        "reference_base0_start_inclusive",
        "reference_base0_end_exclusive",
        "read_base0_start_inclusive",
        "read_base0_end_exclusive"
    ]

    def __init__(
            self,
            name,
            sequence,
            reference_positions,
            quality_scores,
            reference_base0_start_inclusive,
            reference_base0_end_exclusive,
            read_base0_start_inclusive,
            read_base0_end_exclusive):
        """
        Parameters
        ----------
        name : str
            Fragment name, paired reads from the same fragment will have the same name

        sequence : str
            cDNA sequence

        reference_positions : list of (int or None)
            For every base in the sequence, which base-1 reference position
            does it map to, or None if the read base is an insertion or soft-clipped

        quality_scores : array of int
            Base qualities for every character in the sequence

        reference_base0_start_inclusive : int
            Start index of reference locus which is overlapped
            by this read (base 0, inclusive)

        reference_base0_end_exclusive : int
            End index of reference locus which is overlapped
            by this read (base 0, exclusive)

        read_base0_start_inclusive : int or None
            Start index of base in read which corresponds to
            start of reference locus (if it's mapped)

        read_base0_end_exclusive : int or None
            End index after last base in sequence which
            corresponds to reference locus (if it's mapped)
        """
        ######################################################################
        # When can the start or end of the read interval be None?
        # ---------------------------------------------------------
        # If a locus goes from  [x:y) on the reference chromosome then there are
        # a few possibilities for whether x and y will have corresponding mapped
        # positions on a read.
        #
        # If x=y, then we expect the read to either match the reference or have some bases
        # inserted between x and y. In this case, the only reason why x and y wouldn't be mapped
        # is if they occur after the end of a read, but the read still overlaps some
        # part of the interval.
        #
        # If y > x, then we're selecting some non-zero reference bases and either matching them
        # or deleting them. If they CIGAR operation is M (match), then x and y should have
        # corresponding positions on the read (unless, like previously, only part of the
        # interval is covered by a read).
        #
        # In the case of a deletion, however, the selected reference bases do not have
        # corresponding positions on the read.
        #
        # Diagram of a deletion:
        #
        #    REFERENCE
        #    19124 19125 19126 19127 19128 19129 19130 19131 19132 19133 19134
        #      A     C     T     G     G     C     A     T      T    T     T
        #
        # If we're interested in checking whether the sequence 'GCA' is deleted
        # between 19128:19131 then we'll look at an RNA read.
        #
        #    RNA WHICH SUPPORTS REFERENCE
        #    00024 00025 00026 00027 00028 00029 00030 00031 00032 00033 00034
        #      A     C     T     G     G     C     A     T      T    T     T
        #
        #    RNA WHICH SUPPORTS MUTATION
        #    00024 00025 00026 00027 00028 00029 00030 00031 00032 00033 00034
        #      A     C     T     G     T      T    T      T    A     A     A
        #
        # In the RNA read which does not support the mutation (matches reference) the
        # reference index 19128 is mapped to 28 and 19131 is mapped to 31.
        #
        # On the mutant RNA, however, the position 19128, 19129, 19130 are unmapped.
        #
        #    19124 19125 19126 19127 19128 19129 19130 19131 19132 19133 19134
        #      A     C     T     G     G     C     A     T      T    T     T
        #      |     |     |     |                       |      |    |     |
        #      |     |     |     |     *-----------------*      |    |     |
        #      |     |     |     |     |      *-----------------*    |     |
        #      |     |     |     |     |      |    |-----------------*     |
        #      |     |     |     |     |      |    |      *----------------*
        #      |     |     |     |     |      |    |      |
        #    00024 00025 00026 00027 00028 00029 00030 00031 00032 00033 00034
        #      A     C     T     G     T      T    T      T    A     A     A

        self.name = name
        self.sequence = sequence
        self.reference_positions = reference_positions
        self.quality_scores = quality_scores
        self.reference_base0_start_inclusive = reference_base0_start_inclusive
        self.reference_base0_end_exclusive = reference_base0_end_exclusive
        self.read_base0_start_inclusive = read_base0_start_inclusive
        self.read_base0_end_exclusive = read_base0_end_exclusive





from isovar import run_isovar, isovar_results_to_dataframe
from nose.tools import eq_
from testing_helpers import data_path

def test_isovar_main_to_dataframe():
    results = run_isovar(
        variants=data_path("data/b16.f10/b16.vcf"),
        alignment_file=data_path("data/b16.f10/b16.combined.sorted.bam"))
    df = isovar_results_to_dataframe(results)
    print(df)
    eq_(len(df), 4)
    # B16 test data has 2/4 variants with enough coverage
    # to translate protein sequences
    eq_(df["passes_all_filters"].sum(), 2)




# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from collections import defaultdict, Counter

from .default_parameters import MIN_SHARED_FRAGMENTS_FOR_PHASING


def create_variant_to_alt_read_names_dict(isovar_results):
    """
    Create dictionary from variant to names of alt reads supporting
    that variant in an IsovarResult.

    Parameters
    ----------
    isovar_results : list of IsovarResult

    Returns
    -------
    Dictionary from varcode.Variant to set(str) of read names

    """
    return {
        isovar_result.variant: set(isovar_result.alt_read_names)
        for isovar_result in isovar_results
    }


def create_variant_to_protein_sequence_read_names_dict(isovar_results):
    """
    Create dictionary from variant to names of alt reads used to create
    mutant protein sequence in an IsovarResult.

    Parameters
    ----------
    isovar_results : list of IsovarResult

    Returns
    -------
    Dictionary from varcode.Variant to set(str) of read names
    """
    variant_to_read_names = {}
    for isovar_result in isovar_results:
        if isovar_result.has_mutant_protein_sequence_from_rna:
            protein_sequence = isovar_result.top_protein_sequence
            read_names = set(
                protein_sequence.read_names_supporting_protein_sequence)
        else:
            read_names = set()
        variant_to_read_names[isovar_result.variant] = read_names
    return variant_to_read_names


def compute_phasing_counts(variant_to_read_names_dict):
    """

    Parameters
    ----------
    variants_to_read_names : dict
        Dictionary mapping varcode.Variant to set of read names

    Returns
    -------
    Dictionary from variant to Counter(Variant)
    """
    read_names_to_variants = defaultdict(set)

    for variant, read_names in variant_to_read_names_dict.items():
        for read_name in read_names:
            read_names_to_variants[read_name].add(variant)

    # now count up how many reads are shared between pairs of variants
    phasing_counts = defaultdict(Counter)
    for variant, read_names in variant_to_read_names_dict.items():
        for read_name in read_names:
            for other_variant in read_names_to_variants[read_name]:
                if variant != other_variant:
                    phasing_counts[variant][other_variant] += 1
    return phasing_counts

def threshold_phased_variant_counts(counts_dict, min_count):
    """
    Choose set of phased variants by keeping any variants with associated
    counts greater than or equal the given threshold.

    Parameters
    ----------
    counts_dict : variant -> int dict

    min_count : int

    Returns
    -------
    set of varcode.Variant
    """
    return {
        variant
        for (variant, count)
        in counts_dict.items()
        if count >= min_count
    }


def annotate_phased_variants(
        unphased_isovar_results,
        min_shared_fragments_for_phasing=MIN_SHARED_FRAGMENTS_FOR_PHASING):
    """
    Annotate IsovarResult objects with phasing information. Phasing
    is determined by looking at RNA fragments used for assembled protein
    sequences and counting the number of shared fragments.

    Parameters
    ----------
    unphased_isovar_results : list of IsovarResult

    min_shared_fragments_for_phasing : int

    Returns
    -------
    list of IsovarResult
    """

    # create dictionary counting how often each variant co-occurs with others
    # in any reads supporting those variants
    phasing_counts_from_supporting_reads = compute_phasing_counts(
        create_variant_to_alt_read_names_dict(unphased_isovar_results))

    # create dictionary counting how often each variant co-occurs with others
    # in any reads used to construct their protein sequences
    phasing_counts_from_protein_sequences = compute_phasing_counts(
        create_variant_to_protein_sequence_read_names_dict(
            unphased_isovar_results))

    results_with_phasing = []
    for isovar_result in unphased_isovar_results:
        variant = isovar_result.variant
        phased_variants_in_supporting_reads = threshold_phased_variant_counts(
            phasing_counts_from_supporting_reads[variant],
            min_count=min_shared_fragments_for_phasing)
        phased_variants_in_protein_sequence = \
            threshold_phased_variant_counts(
                phasing_counts_from_protein_sequences[variant],
                min_count=min_shared_fragments_for_phasing)
        results_with_phasing.append(
            isovar_result.clone_with_updates(
                phased_variants_in_supporting_reads=phased_variants_in_supporting_reads,
                phased_variants_in_protein_sequence=phased_variants_in_protein_sequence))
    return results_with_phasing



from isovar.alignment_score import alignment_score
from nose.tools import eq_

def test_alignment_score_same_strings():
    eq_(alignment_score("", ""), 0)
    eq_(alignment_score("a", "a"), 0)
    eq_(alignment_score("WUZZLE", "WUZZLE"), 0)

def test_alignment_score_strings_of_different_length():
    eq_(alignment_score("ab", "a"), 1)
    eq_(alignment_score("a", "ab"), 1)
    eq_(alignment_score("WUZZLE", "WUZZLE?"), 1)
    eq_(alignment_score("SNUZZLE", "UZZLE"), 2)


def test_alignment_score_totally_different_strings():
    eq_(alignment_score("", "a"), 1)
    eq_(alignment_score("", "ab"), 2)
    eq_(alignment_score("WUZZLE", "HEAVY"), 6)
    eq_(alignment_score("DOG", "CATCATCAT"), 9)
    eq_(alignment_score("a", ""), 1)
    eq_(alignment_score("ab", ""), 2)
    eq_(alignment_score("WUZZLE", "HEAVY"), 6)
    eq_(alignment_score("CATCATCAT", "DOG"), 9)


def test_alignment_min_subsequence_length():
    # if matching subsequence isn't long enough then misalignment score
    # will be sum of two sequence lengths
    eq_(alignment_score("aaa", "aaa", min_subsequence_length=10), 6)



# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module combines variant cDNA sequences collected from a BAM file with
the reading frames of annotated reference transcripts to create candidate
translations.
"""


from __future__ import print_function, division, absolute_import

from .logging import get_logger
from .translation_key import TranslationKey


logger = get_logger(__name__)


class Translation(TranslationKey):
    """
    Translated amino acid sequence of a VariantORF for a particular
    ReferenceContext and VariantSequence.
    """
    __slots__ = [
        "untrimmed_variant_sequence",
        "reference_context",
        "variant_orf"
    ]

    def __init__(
            self,
            amino_acids,
            variant_aa_interval_start,
            variant_aa_interval_end,
            ends_with_stop_codon,
            frameshift,
            untrimmed_variant_sequence,
            reference_context,
            variant_orf):
        # TODO:
        #  get rid of untrimmed_variant_sequence by making
        #  VariantORF keep track of its inputs
        self.amino_acids = amino_acids
        self.variant_aa_interval_start = variant_aa_interval_start
        self.variant_aa_interval_end = variant_aa_interval_end
        self.ends_with_stop_codon = ends_with_stop_codon
        self.frameshift = frameshift
        # this variant sequence might differ from the one
        # in variant_orf due to trimming required to match the reference
        self.untrimmed_variant_sequence = untrimmed_variant_sequence
        self.reference_context = reference_context
        self.variant_orf = variant_orf

    @property
    def reads(self):
        """
        RNA reads which were used to construct the coding sequence
        from which we translated these amino acids.
        """
        return self.untrimmed_variant_sequence.reads

    @property
    def reference_cdna_sequence_before_variant(self):
        """

        Returns str
        """
        return (
            self.variant_orf.reference_cdna_sequence_before_variant)


    @property
    def num_mismatches_before_variant(self):
        """
        Number of nucleotides in the variant cDNA sequence which
        don't match the ReferenceContext transcript sequence at
        positions before the variant locus.

        Returns int
        """
        return self.variant_orf.num_mismatches_before_variant

    @property
    def num_mismatches_after_variant(self):
        """
        Number of nucleotides in the variant cDNA sequence which
        don't match the ReferenceContext transcript sequence at
        positions after the variant locus.

        Returns int
        """
        return self.variant_orf.num_mismatches_after_variant

    @property
    def cdna_sequence(self):
        """
        cDNA sequence assembled from variant supporting reads

        Returns str
        """
        return self.variant_orf.cdna_sequence

    @property
    def offset_to_first_complete_codon(self):
        """
        Offset to first complete codon in the cDNA sequence

        Returns int in {0, 1, 2}
        """
        return self.variant_orf.offset_to_first_complete_codon

    @property
    def variant_cdna_interval_start(self):
        """
        Interbase start coordinate of variant interval in the cDNA sequence
        """
        return self.variant_orf.variant_cdna_interval_start

    @property
    def variant_cdna_interval_end(self):
        """
        Interbase end coordinate of variant interval in the cDNA sequence
        """
        return self.variant_orf.variant_cdna_interval_end

    def as_translation_key(self):
        """
        Project Translation object or any other derived class into just a
        TranslationKey, which has fewer fields and can be used as a
        dictionary key.
        """
        return TranslationKey(**{
            name: getattr(self, name)
            for name in TranslationKey._fields})





# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from .main_args import make_isovar_arg_parser, run_isovar_from_parsed_args
from .protein_sequence_args import protein_sequence_creator_from_args
from .rna_args import read_collector_from_args

__all__ = [
    "make_isovar_arg_parser",
    "run_isovar_from_parsed_args",
    "protein_sequence_creator_from_args",
    "read_collector_from_args",
]


# Copyright (c) 2016-2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

__version__ = "1.0.10"


from .allele_read import AlleleRead
from .dataframe_helpers import isovar_results_to_dataframe
from .isovar_result import IsovarResult
from .locus_read import LocusRead
from .main import run_isovar
from .protein_sequence import ProteinSequence
from .protein_sequence_creator import ProteinSequenceCreator
from .read_collector import ReadCollector
from .read_evidence import ReadEvidence
from .variant_orf import VariantORF
from .variant_sequence import VariantSequence
from .variant_sequence_creator import VariantSequenceCreator


__all__ = [
    "run_isovar",
    "isovar_results_to_dataframe",
    "AlleleRead",
    "IsovarResult",
    "LocusRead",
    "ProteinSequence",
    "ProteinSequenceCreator",
    "ReadCollector",
    "ReadEvidence",
    "VariantORF",
    "VariantSequence",
    "VariantSequenceCreator",
]



# Copyright (c) 2019. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common command-line arguments for filtering Isovar results
"""

from __future__ import print_function, division, absolute_import

from collections import OrderedDict

from ..default_parameters import (
    MIN_NUM_RNA_ALT_FRAGMENTS,
    MIN_NUM_RNA_ALT_READS,
    MIN_FRACTION_RNA_ALT_FRAGMENTS,
    MIN_RATIO_RNA_ALT_TO_OTHER_FRAGMENTS
)

def add_filter_args(parser):
    """
    Extends an ArgumentParser instance with commandline arguments related
    to filtering variants and/or their associated protein sequences.
    """
    filter_group = parser.add_argument_group("Filtering")
    filter_group.add_argument(
        "--min-alt-rna-reads",
        type=int,
        default=MIN_NUM_RNA_ALT_READS,
        help="Minimum number of reads supporting variant allele (default %(default)s)")

    filter_group.add_argument(
        "--min-alt-rna-fragments",
        type=int,
        default=MIN_NUM_RNA_ALT_FRAGMENTS,
        help=(
            "Minimum number of fragments supporting variant allele (default %(default)s). "
            "Note that this option is the same as --min-alt-rna-reads for single-end "
            "sequencing."))

    filter_group.add_argument(
        "--min-alt-rna-fraction",
        type=float,
        default=MIN_FRACTION_RNA_ALT_FRAGMENTS,
        help=(
            "Minimum ratio of fragments supporting variant allele to total RNA fragments "
            "(default %(default)s)."))

    filter_group.add_argument(
        "--min-ratio-alt-to-other-fragments",
        type=float,
        default=MIN_RATIO_RNA_ALT_TO_OTHER_FRAGMENTS,
        help=(
            "At loci where alleles other than the ref and a single alt are supported, "
            "this parameter controls how many more times fragments supporting "
            "the variant allele are required relative to other non-reference "
            "alleles (default %(default)s)."))
    return filter_group


def filter_threshold_dict_from_args(args):
    """
    Convert names of filters from external CLI options to more
    consistent naming scheme of {min|max}_{Isovar property}.

    Returns OrderedDict
    """
    d = OrderedDict()
    d["min_ratio_alt_to_other_fragments"] = args.min_ratio_alt_to_other_fragments
    d["min_fraction_alt_fragments"] = args.min_alt_rna_fraction
    d["min_num_alt_fragments"] = args.min_alt_rna_fragments
    d["min_num_alt_reads"] = args.min_alt_rna_reads
    return d



from isovar.effect_prediction import (
    top_varcode_effect,
    reference_coding_transcripts_for_variant
)
from varcode import Variant
from varcode.effects import Intergenic
from nose.tools import eq_

# intergenic variant from error log of using Isovar 1.0.0
intergenic_variant = Variant('1', 30256419, 'G', 'T', 'GRCh37')


def test_top_effect_outside_of_gene():
    top_effect = top_varcode_effect(intergenic_variant)
    assert top_effect is not None
    assert isinstance(top_effect, Intergenic)


def test_reference_coding_transcripts_outside_of_gene():
    transcripts = \
        reference_coding_transcripts_for_variant(
            intergenic_variant)
    eq_(len(transcripts), 0)

